import asyncio
import inspect
import socket
import threading
import time
import unittest

from .. import Hermes, backend


def load_tests(loader, tests, pattern):
  from . import abstract, facade, inprocess, memcached, redis  # noqa: F401

  suite = unittest.TestSuite()
  for m in filter(inspect.ismodule, locals().values()):
    suite.addTests(loader.loadTestsFromModule(m))

  return suite


class TestCase(unittest.TestCase):

  testee = None
  fixture = None

  def _arghash(self, *args, **kwargs):
    '''
    Not very neat as it penetrates into an implementation detail,
    though otherwise it'll be harder to make assertion on keys.
    '''

    arguments = args, tuple(sorted(kwargs.items()))
    return self.testee.mangler.hash(self.testee.mangler.dumps(arguments))


def createFixture(cache):  # noqa: C901

  class Fixture:

    calls = 0

    @cache
    def simple(self, a, b):
      '''Here be dragons... seriously just a docstring test.'''

      self.calls += 1
      return '{0}+{1}'.format(a, b)[::-1]

    @cache
    def nested(self, a, b):
      self.calls += 1
      return self.simple(b, a)[::-1]

    @cache(tags = ('rock', 'tree'))
    def tagged(self, a, b):
      self.calls += 1
      return '{0}-{1}'.format(a, b)[::-2]

    @cache(tags = ('rock', 'ice'))
    def tagged2(self, a, b):
      self.calls += 1
      return '{0}%{1}'.format(a, b)[::-2]

    @cache(tags = ('ash', 'stone'), key = lambda fn, a, b: f'mykey:{a}:{b}')
    def key(self, a, b):
      self.calls += 1
      return '{0}*{1}'.format(a, b)[::2]

    @cache(
      tags = ('a', 'z'),
      key = lambda fn, *a: 'mk:{0}:{1}'.format(*a).replace(' ', ''),
      ttl = 1200,
    )
    def all(self, a, b):
      self.calls += 1
      return {'a': a['alpha'], 'b': {'b': b[0]}}

    @cache
    async def asimple(self, a, b):
      await asyncio.sleep(0)
      self.calls += 1
      return '{0}+{1}'.format(a, b)[::-1]

    @cache
    async def anested(self, a, b):
      self.calls += 1
      return (await self.asimple(b, a))[::-1]

    @cache(tags = ('rock', 'tree'))
    async def atagged(self, a, b):
      await asyncio.sleep(0)
      self.calls += 1
      return '{0}-{1}'.format(a, b)[::-2]

    @cache(tags = ('rock', 'ice'))
    async def atagged2(self, a, b):
      await asyncio.sleep(0)
      self.calls += 1
      return '{0}%{1}'.format(a, b)[::-2]

    @cache(tags = ('ash', 'stone'), key = lambda fn, a, b: f'mykey:{a}:{b}')
    async def akey(self, a, b):
      await asyncio.sleep(0)
      self.calls += 1
      return '{0}*{1}'.format(a, b)[::2]

    @cache(
      tags = ('a', 'z'),
      key = lambda fn, *a: 'mk:{0}:{1}'.format(*a).replace(' ', ''),
      ttl = 1200,
    )
    async def aall(self, a, b):
      await asyncio.sleep(0)
      self.calls += 1
      return {'a': a['alpha'], 'b': {'b': b[0]}}

  return Fixture()


def asynctest(coro):
  assert inspect.iscoroutinefunction(coro)

  def wrapper(self):
    return asyncio.run(coro(self))

  return wrapper


class FakeBackendServer:

  port = None
  '''Fake server port.'''

  log = None
  '''Activity log.'''

  thread = None
  '''Server connection-accepting thread.'''

  closing = None
  '''Whether connection is closing.'''

  def __init__(self):
    self.log = []

  def serve(self):
    self.closing = False

    serverSocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    serverSocket.bind(('', 0))
    serverSocket.listen(1)

    self.port = serverSocket.getsockname()[1]

    self.thread = threading.Thread(target = self.target, args = (serverSocket,))
    self.thread.start()

  def close(self):
    self.closing = True

    clientSock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    clientSock.connect(('localhost', self.port))

    clientSock.close()

    self.thread.join()

  def target(self, serverSocket):
    clientSocket, _ = serverSocket.accept()

    if not self.closing:
      self.log.append('connected')

    try:
      chunk = clientSocket.recv(1024)
      if not self.closing:
        if not chunk:
          self.log.append('closed')
        else:
          self.log.append('received {}'.format(chunk))
    finally:
      clientSocket.close()
      serverSocket.close()


class BaseRemoteBackendTest(TestCase):

  ttlDelta = 0

  def setUp(self):
    '''``testee`` is expected to be created in override.'''

    self.fixture = createFixture(self.testee)

    self.testee.clean()
    self.addCleanup(self.testee.clean)

  def getSize(self) -> int:
    raise NotImplementedError

  def getKeyTtl(self, key: str) -> int:
    raise NotImplementedError

  def assertEmptyBackend(self):
    self.assertEqual(0, self.getSize())

  def testSimple(self):
    self.assertEqual(0, self.fixture.calls)
    self.assertEmptyBackend()

    key = 'cache:entry:hermes.test:Fixture:simple:' + self._arghash('alpha', b = 'beta')
    for _ in range(4):
      self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', b = 'beta'))
      self.assertEqual(1, self.fixture.calls)
      self.assertEqual(1, self.getSize())

      self.assertAlmostEqual(360, self.getKeyTtl(key), delta = self.ttlDelta)
      self.assertEqual(
        'ateb+ahpla', self.testee.mangler.loads(self.testee.backend.client.get(key))
      )

    self.fixture.simple.invalidate('alpha', b = 'beta')
    self.assertEmptyBackend()


    expected = "]}'ammag'{[+}]'ateb'[ :'ahpla'{"
    key      = 'cache:entry:hermes.test:Fixture:simple:{}'.format(
      self._arghash({'alpha': ['beta']}, [{'gamma'}]))
    for _ in range(4):
      self.assertEqual(expected, self.fixture.simple({'alpha': ['beta']}, [{'gamma'}]))

      self.assertEqual(2, self.fixture.calls)
      self.assertEqual(1, self.getSize())

      self.assertAlmostEqual(360, self.getKeyTtl(key), delta = self.ttlDelta)

      self.assertEqual(expected, self.testee.mangler.loads(self.testee.backend.client.get(key)))

  @asynctest
  async def testAsyncSimple(self):
    self.assertEqual(0, self.fixture.calls)
    self.assertEmptyBackend()

    key = 'cache:entry:hermes.test:Fixture:asimple:' + self._arghash('alpha', b = 'beta')
    for _ in range(4):
      self.assertEqual('ateb+ahpla', await self.fixture.asimple('alpha', b = 'beta'))
      self.assertEqual(1, self.fixture.calls)
      self.assertEqual(1, self.getSize())

      self.assertAlmostEqual(360, self.getKeyTtl(key), delta = self.ttlDelta)
      self.assertEqual(
        'ateb+ahpla', self.testee.mangler.loads(self.testee.backend.client.get(key))
      )

    await self.fixture.asimple.invalidate('alpha', b = 'beta')
    self.assertEmptyBackend()


    expected = "]}'ammag'{[+}]'ateb'[ :'ahpla'{"
    key      = 'cache:entry:hermes.test:Fixture:asimple:{}'.format(
      self._arghash({'alpha': ['beta']}, [{'gamma'}]))
    for _ in range(4):
      self.assertEqual(expected, await self.fixture.asimple({'alpha': ['beta']}, [{'gamma'}]))

      self.assertEqual(2, self.fixture.calls)
      self.assertEqual(1, self.getSize())

      self.assertAlmostEqual(360, self.getKeyTtl(key), delta = self.ttlDelta)

      self.assertEqual(expected, self.testee.mangler.loads(self.testee.backend.client.get(key)))

  def testTagged(self):
    self.assertEqual(0, self.fixture.calls)
    self.assertEmptyBackend()

    for _ in range(4):
      self.assertEqual('ae-hl', self.fixture.tagged('alpha', b = 'beta'))
      self.assertEqual(1,       self.fixture.calls)
      self.assertEqual(3,       self.getSize())

      self.assertEqual(-1,  self.getKeyTtl('cache:tag:rock'))
      self.assertEqual(-1,  self.getKeyTtl('cache:tag:tree'))
      rockTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:rock'))
      treeTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:tree'))
      self.assertFalse(rockTag == treeTag)
      self.assertEqual(22, len(rockTag))
      self.assertEqual(22, len(treeTag))

      argHash = self._arghash('alpha', b = 'beta')
      tagHash = self.testee.mangler.hashTags(dict(tree = treeTag, rock = rockTag))
      key     = 'cache:entry:hermes.test:Fixture:tagged:{0}:{1}'.format(argHash, tagHash)
      self.assertAlmostEqual(360, self.getKeyTtl(key), delta = self.ttlDelta)
      self.assertEqual('ae-hl', self.testee.mangler.loads(self.testee.backend.client.get(key)))

    self.fixture.tagged.invalidate('alpha', b = 'beta')

    self.assertEqual(2, self.getSize())

    rockTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:rock'))
    treeTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:tree'))
    self.assertNotEqual(rockTag, treeTag)
    self.assertEqual(22, len(rockTag))
    self.assertEqual(22, len(treeTag))

    for _ in range(4):
      self.assertEqual('ae-hl', self.fixture.tagged('alpha', 'beta'))
      self.assertEqual('ae%hl', self.fixture.tagged2('alpha', 'beta'))
      self.assertEqual(3, self.fixture.calls)

      self.assertEqual(5, self.getSize())
      self.assertEqual(
        rockTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:rock'))
      )
      self.assertEqual(
        treeTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:tree'))
      )
      self.assertEqual(
        22, len(self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:ice')))
      )

    self.testee.clean(['rock'])

    self.assertEqual(4, self.getSize())
    self.assertIsNone(self.testee.backend.client.get('cache:tag:rock'))
    iceTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:ice'))

    for _ in range(4):
      self.assertEqual('ae-hl', self.fixture.tagged('alpha', 'beta'))
      self.assertEqual('ae%hl', self.fixture.tagged2('alpha', 'beta'))
      self.assertEqual(5, self.fixture.calls)

      size = self.getSize()
      self.assertEqual(7, size, 'has new and old entries for tagged and tagged 2 + 3 tags')
      self.assertEqual(
        treeTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:tree'))
      )
      self.assertEqual(
        iceTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:ice'))
      )
      self.assertEqual(
        22, len(self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:rock')))
      )
      self.assertNotEqual(
        rockTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:rock'))
      )

  @asynctest
  async def testAsyncTagged(self):
    self.assertEqual(0, self.fixture.calls)
    self.assertEmptyBackend()

    for _ in range(4):
      self.assertEqual('ae-hl', await self.fixture.atagged('alpha', b = 'beta'))
      self.assertEqual(1,       self.fixture.calls)
      self.assertEqual(3,       self.getSize())

      self.assertEqual(-1,  self.getKeyTtl('cache:tag:rock'))
      self.assertEqual(-1,  self.getKeyTtl('cache:tag:tree'))
      rockTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:rock'))
      treeTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:tree'))
      self.assertFalse(rockTag == treeTag)
      self.assertEqual(22, len(rockTag))
      self.assertEqual(22, len(treeTag))

      argHash = self._arghash('alpha', b = 'beta')
      tagHash = self.testee.mangler.hashTags(dict(tree = treeTag, rock = rockTag))
      key     = 'cache:entry:hermes.test:Fixture:atagged:{0}:{1}'.format(argHash, tagHash)
      self.assertAlmostEqual(360, self.getKeyTtl(key), delta = self.ttlDelta)
      self.assertEqual('ae-hl', self.testee.mangler.loads(self.testee.backend.client.get(key)))

    await self.fixture.atagged.invalidate('alpha', b = 'beta')

    self.assertEqual(2, self.getSize())

    rockTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:rock'))
    treeTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:tree'))
    self.assertNotEqual(rockTag, treeTag)
    self.assertEqual(22, len(rockTag))
    self.assertEqual(22, len(treeTag))

    for _ in range(4):
      self.assertEqual('ae-hl', await self.fixture.atagged('alpha', 'beta'))
      self.assertEqual('ae%hl', await self.fixture.atagged2('alpha', 'beta'))
      self.assertEqual(3, self.fixture.calls)

      self.assertEqual(5, self.getSize())
      self.assertEqual(
        rockTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:rock'))
      )
      self.assertEqual(
        treeTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:tree'))
      )
      self.assertEqual(
        22, len(self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:ice')))
      )

    self.testee.clean(['rock'])

    self.assertEqual(4, self.getSize())
    self.assertIsNone(self.testee.backend.client.get('cache:tag:rock'))
    iceTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:ice'))

    for _ in range(4):
      self.assertEqual('ae-hl', self.fixture.tagged('alpha', 'beta'))
      self.assertEqual('ae%hl', self.fixture.tagged2('alpha', 'beta'))
      self.assertEqual(5, self.fixture.calls)

      size = self.getSize()
      self.assertEqual(7, size, 'has new and old entries for tagged and tagged 2 + 3 tags')
      self.assertEqual(
        treeTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:tree'))
      )
      self.assertEqual(
        iceTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:ice'))
      )
      self.assertEqual(
        22, len(self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:rock')))
      )
      self.assertNotEqual(
        rockTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:rock'))
      )

  def testFunction(self):
    counter = dict(foo = 0, bar = 0)

    @self.testee
    def foo(a, b):
      counter['foo'] += 1
      return '{0}+{1}'.format(a, b)[::-1]

    key = lambda fn, *args, **kwargs: 'mk:{0}:{1}'.format(*args)

    @self.testee(tags = ('a', 'z'), key = key, ttl = 120)
    def bar(a, b):
      counter['bar'] += 1
      return '{0}-{1}'.format(a, b)[::2]


    self.assertEqual(0, counter['foo'])
    self.assertEmptyBackend()

    key = 'cache:entry:hermes.test:foo:' + self._arghash('alpha', 'beta')
    for _ in range(4):
      self.assertEqual('ateb+ahpla', foo('alpha', 'beta'))

      self.assertEqual(1, counter['foo'])
      self.assertEqual(1, self.getSize())

      self.assertAlmostEqual(360, self.getKeyTtl(key), delta = self.ttlDelta)
      self.assertEqual(
        'ateb+ahpla', self.testee.mangler.loads(self.testee.backend.client.get(key))
      )

    foo.invalidate('alpha', 'beta')
    self.assertFalse(self.testee.backend.client.get(key))
    self.assertEmptyBackend()


    self.assertEqual(0, counter['bar'])
    self.assertEmptyBackend()

    for _ in range(4):
      self.assertEqual('apabt', bar('alpha', 'beta'))
      self.assertEqual(1,       counter['bar'])
      self.assertEqual(3,       self.getSize())

      self.assertEqual(-1, self.getKeyTtl('cache:tag:a'))
      self.assertEqual(-1, self.getKeyTtl('cache:tag:z'))

      aTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:a'))
      zTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:z'))
      self.assertFalse(aTag == zTag)
      self.assertEqual(22, len(aTag))
      self.assertEqual(22, len(zTag))

      tagHash = self.testee.mangler.hashTags(dict(a = aTag, z = zTag))
      key     = 'mk:alpha:beta:' + tagHash
      self.assertAlmostEqual(120, self.getKeyTtl(key), delta = self.ttlDelta)
      self.assertEqual('apabt', self.testee.mangler.loads(self.testee.backend.client.get(key)))

    bar.invalidate('alpha', 'beta')
    self.assertEqual(1,  counter['foo'])
    self.assertIsNone(self.testee.backend.client.get(key))
    self.assertEqual(2, self.getSize())

    self.assertEqual('apabt', bar('alpha', 'beta'))
    self.assertEqual(2, counter['bar'])
    self.assertEqual(3, self.getSize())

  @asynctest
  async def testAsyncFunction(self):
    counter = dict(foo = 0, bar = 0)

    @self.testee
    async def foo(a, b):
      await asyncio.sleep(0)
      counter['foo'] += 1
      return '{0}+{1}'.format(a, b)[::-1]

    key = lambda fn, *args, **kwargs: 'mk:{0}:{1}'.format(*args)

    @self.testee(tags = ('a', 'z'), key = key, ttl = 120)
    async def bar(a, b):
      counter['bar'] += 1
      return '{0}-{1}'.format(a, b)[::2]


    self.assertEqual(0, counter['foo'])
    self.assertEmptyBackend()

    key = 'cache:entry:hermes.test:foo:' + self._arghash('alpha', 'beta')
    for _ in range(4):
      self.assertEqual('ateb+ahpla', await foo('alpha', 'beta'))

      self.assertEqual(1, counter['foo'])
      self.assertEqual(1, self.getSize())

      self.assertAlmostEqual(360, self.getKeyTtl(key), delta = self.ttlDelta)
      self.assertEqual(
        'ateb+ahpla', self.testee.mangler.loads(self.testee.backend.client.get(key))
      )

    await foo.invalidate('alpha', 'beta')
    self.assertFalse(self.testee.backend.client.get(key))
    self.assertEmptyBackend()


    self.assertEqual(0, counter['bar'])
    self.assertEmptyBackend()

    for _ in range(4):
      self.assertEqual('apabt', await bar('alpha', 'beta'))
      self.assertEqual(1, counter['bar'])
      self.assertEqual(3, self.getSize())

      self.assertEqual(-1, self.getKeyTtl('cache:tag:a'))
      self.assertEqual(-1, self.getKeyTtl('cache:tag:z'))

      aTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:a'))
      zTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:z'))
      self.assertFalse(aTag == zTag)
      self.assertEqual(22, len(aTag))
      self.assertEqual(22, len(zTag))

      tagHash = self.testee.mangler.hashTags(dict(a = aTag, z = zTag))
      key     = 'mk:alpha:beta:' + tagHash
      self.assertAlmostEqual(120, self.getKeyTtl(key), delta = self.ttlDelta)
      self.assertEqual('apabt', self.testee.mangler.loads(self.testee.backend.client.get(key)))

    await bar.invalidate('alpha', 'beta')
    self.assertEqual(1,  counter['foo'])
    self.assertIsNone(self.testee.backend.client.get(key))
    self.assertEqual(2, self.getSize())

    self.assertEqual('apabt', await bar('alpha', 'beta'))
    self.assertEqual(2, counter['bar'])
    self.assertEqual(3, self.getSize())

  def testKey(self):
    self.assertEqual(0, self.fixture.calls)
    self.assertEmptyBackend()

    for _ in range(4):
      self.assertEqual('apabt', self.fixture.key('alpha', 'beta'))
      self.assertEqual(1,       self.fixture.calls)
      self.assertEqual(3,       self.getSize())

      self.assertEqual(-1,  self.getKeyTtl('cache:tag:ash'))
      self.assertEqual(-1,  self.getKeyTtl('cache:tag:stone'))
      ashTag   = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:ash'))
      stoneTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:stone'))
      self.assertFalse(ashTag == stoneTag)
      self.assertEqual(22, len(ashTag))
      self.assertEqual(22, len(stoneTag))

      tagHash = self.testee.mangler.hashTags(dict(ash = ashTag, stone = stoneTag))
      key     = 'mykey:alpha:beta:' + tagHash
      self.assertAlmostEqual(360, self.getKeyTtl(key), delta = self.ttlDelta)
      self.assertEqual('apabt', self.testee.mangler.loads(self.testee.backend.client.get(key)))

    self.fixture.key.invalidate('alpha', 'beta')

    self.assertIsNone(self.testee.backend.client.get(key))
    self.assertEqual(2, self.getSize())

    self.assertEqual(
      ashTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:ash'))
    )
    self.assertEqual(
      stoneTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:stone'))
    )

  @asynctest
  async def testAsyncKey(self):
    self.assertEqual(0, self.fixture.calls)
    self.assertEmptyBackend()

    for _ in range(4):
      self.assertEqual('apabt', await self.fixture.akey('alpha', 'beta'))
      self.assertEqual(1,       self.fixture.calls)
      self.assertEqual(3,       self.getSize())

      self.assertEqual(-1,  self.getKeyTtl('cache:tag:ash'))
      self.assertEqual(-1,  self.getKeyTtl('cache:tag:stone'))
      ashTag   = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:ash'))
      stoneTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:stone'))
      self.assertFalse(ashTag == stoneTag)
      self.assertEqual(22, len(ashTag))
      self.assertEqual(22, len(stoneTag))

      tagHash = self.testee.mangler.hashTags(dict(ash = ashTag, stone = stoneTag))
      key     = 'mykey:alpha:beta:' + tagHash
      self.assertAlmostEqual(360, self.getKeyTtl(key), delta = self.ttlDelta)
      self.assertEqual('apabt', self.testee.mangler.loads(self.testee.backend.client.get(key)))

    await self.fixture.akey.invalidate('alpha', 'beta')

    self.assertIsNone(self.testee.backend.client.get(key))
    self.assertEqual(2, self.getSize())

    self.assertEqual(
      ashTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:ash'))
    )
    self.assertEqual(
      stoneTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:stone'))
    )

  def testAll(self):
    self.assertEqual(0, self.fixture.calls)
    self.assertEmptyBackend()
    for _ in range(4):
      self.assertEqual({'a': 1, 'b': {'b': 'beta'}}, self.fixture.all({'alpha': 1}, ['beta']))
      self.assertEqual(1, self.fixture.calls)
      self.assertEqual(3, self.getSize())

      self.assertEqual(-1, self.getKeyTtl('cache:tag:a'))
      self.assertEqual(-1, self.getKeyTtl('cache:tag:z'))

      aTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:a'))
      zTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:z'))
      self.assertFalse(aTag == zTag)
      self.assertEqual(22, len(aTag))
      self.assertEqual(22, len(zTag))

      tagHash = self.testee.mangler.hashTags(dict(a = aTag, z = zTag))
      key = "mk:{'alpha':1}:['beta']:" + tagHash
      self.assertAlmostEqual(1200, self.getKeyTtl(key), delta = self.ttlDelta)
      self.assertEqual(
        {'a': 1, 'b': {'b': 'beta'}},
        self.testee.mangler.loads(self.testee.backend.client.get(key)),
      )

    self.fixture.all.invalidate({'alpha': 1}, ['beta'])

    self.assertEqual(2, self.getSize())
    self.assertEqual(
      aTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:a'))
    )
    self.assertEqual(
      zTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:z'))

    )

  @asynctest
  async def testAsyncAll(self):
    self.assertEqual(0, self.fixture.calls)
    self.assertEmptyBackend()
    for _ in range(4):
      self.assertEqual(
        {'a': 1, 'b': {'b': 'beta'}}, await self.fixture.aall({'alpha': 1}, ['beta'])
      )
      self.assertEqual(1, self.fixture.calls)
      self.assertEqual(3, self.getSize())

      self.assertEqual(-1, self.getKeyTtl('cache:tag:a'))
      self.assertEqual(-1, self.getKeyTtl('cache:tag:z'))

      aTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:a'))
      zTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:z'))
      self.assertFalse(aTag == zTag)
      self.assertEqual(22, len(aTag))
      self.assertEqual(22, len(zTag))

      tagHash = self.testee.mangler.hashTags(dict(a = aTag, z = zTag))
      key = "mk:{'alpha':1}:['beta']:" + tagHash
      self.assertAlmostEqual(1200, self.getKeyTtl(key), delta = self.ttlDelta)
      self.assertEqual(
        {'a': 1, 'b': {'b': 'beta'}},
        self.testee.mangler.loads(self.testee.backend.client.get(key)),
      )

    await self.fixture.aall.invalidate({'alpha': 1}, ['beta'])

    self.assertEqual(2, self.getSize())
    self.assertEqual(
      aTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:a'))
    )
    self.assertEqual(
      zTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:z'))
    )

  def testClean(self):
    self.assertEqual(0, self.fixture.calls)
    self.assertEmptyBackend()

    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(2,            self.fixture.calls)
    self.assertEqual(4,            self.getSize())

    self.testee.clean()

    self.assertEqual(2, self.fixture.calls)
    self.assertEmptyBackend()

  @asynctest
  async def testAsyncClean(self):
    self.assertEqual(0, self.fixture.calls)
    self.assertEmptyBackend()

    self.assertEqual('ateb+ahpla', await self.fixture.asimple('alpha', 'beta'))
    self.assertEqual('aldamg',     await self.fixture.atagged('gamma', 'delta'))
    self.assertEqual(2,            self.fixture.calls)
    self.assertEqual(4,            self.getSize())

    self.testee.clean()

    self.assertEqual(2, self.fixture.calls)
    self.assertEmptyBackend()

  def testCleanTagged(self):
    self.assertEqual(0, self.fixture.calls)
    self.assertEmptyBackend()

    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(2,            self.fixture.calls)
    self.assertEqual(4,            self.getSize())

    simpleKey = 'cache:entry:hermes.test:Fixture:simple:' + self._arghash('alpha', 'beta')
    self.assertEqual(
      'ateb+ahpla', self.testee.mangler.loads(self.testee.backend.client.get(simpleKey))
    )

    rockTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:rock'))
    treeTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:tree'))
    self.assertFalse(rockTag == treeTag)
    self.assertEqual(22, len(rockTag))
    self.assertEqual(22, len(treeTag))

    argHash   = self._arghash('gamma', 'delta')
    tagHash   = self.testee.mangler.hashTags(dict(tree = treeTag, rock = rockTag))
    taggedKey = 'cache:entry:hermes.test:Fixture:tagged:{0}:{1}'.format(argHash, tagHash)
    self.assertEqual(
      'aldamg', self.testee.mangler.loads(self.testee.backend.client.get(taggedKey))
    )


    self.testee.clean(('rock',))
    self.assertEqual(3, self.getSize())

    self.assertEqual(
      'ateb+ahpla', self.testee.mangler.loads(self.testee.backend.client.get(simpleKey))
    )

    self.assertIsNone(self.testee.backend.client.get('cache:tag:rock'))
    self.assertEqual(
      treeTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:tree'))
    )

    # stale still accessible, though only directly
    self.assertEqual(
      'aldamg', self.testee.mangler.loads(self.testee.backend.client.get(taggedKey))
    )


    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(3,            self.fixture.calls)
    self.assertEqual(5,            self.getSize(), '+1 old tagged entry')

    self.assertEqual(
      'ateb+ahpla', self.testee.mangler.loads(self.testee.backend.client.get(simpleKey))
    )

    self.assertNotEqual(
      rockTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:rock'))
    )
    rockTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:rock'))
    self.assertFalse(rockTag == treeTag)
    self.assertEqual(22, len(rockTag))
    self.assertEqual(
      treeTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:tree'))
    )

    # stale still accessible, though only directly
    self.assertEqual(
      'aldamg', self.testee.mangler.loads(self.testee.backend.client.get(taggedKey))
    )

    argHash   = self._arghash('gamma', 'delta')
    tagHash   = self.testee.mangler.hashTags(dict(tree = treeTag, rock = rockTag))
    taggedKey = 'cache:entry:hermes.test:Fixture:tagged:{0}:{1}'.format(argHash, tagHash)
    self.assertEqual(
      'aldamg', self.testee.mangler.loads(self.testee.backend.client.get(taggedKey))
    )


    self.testee.clean(('rock', 'tree'))
    self.assertEqual(3, self.getSize(), 'simaple, new tagged and old tagged')

    self.assertEqual(
      'ateb+ahpla', self.testee.mangler.loads(self.testee.backend.client.get(simpleKey))
    )

    # new stale is accessible, though only directly
    self.assertEqual(
      'aldamg', self.testee.mangler.loads(self.testee.backend.client.get(taggedKey))
    )


    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(4,            self.fixture.calls)
    self.assertEqual(6,            self.getSize(), '+2 old tagged entries')

    self.assertEqual(
      'ateb+ahpla', self.testee.mangler.loads(self.testee.backend.client.get(simpleKey))
    )

    # new stale still accessible, though only directly
    self.assertEqual(
      'aldamg', self.testee.mangler.loads(self.testee.backend.client.get(taggedKey))
    )

    self.assertNotEqual(
      rockTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:rock'))
    )
    self.assertNotEqual(
      treeTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:tree'))
    )

    rockTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:rock'))
    treeTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:tree'))
    self.assertFalse(rockTag == treeTag)
    self.assertEqual(22, len(rockTag))
    self.assertEqual(22, len(treeTag))

    argHash   = self._arghash('gamma', 'delta')
    tagHash   = self.testee.mangler.hashTags(dict(tree = treeTag, rock = rockTag))
    taggedKey = 'cache:entry:hermes.test:Fixture:tagged:{0}:{1}'.format(argHash, tagHash)
    self.assertEqual(
      'aldamg', self.testee.mangler.loads(self.testee.backend.client.get(taggedKey))
    )


    self.testee.clean(('rock', 'tree'))
    self.testee.clean()
    self.assertEmptyBackend()

  @asynctest
  async def testAsyncCleanTagged(self):
    self.assertEqual(0, self.fixture.calls)
    self.assertEmptyBackend()

    self.assertEqual('ateb+ahpla', await self.fixture.asimple('alpha', 'beta'))
    self.assertEqual('aldamg',     await self.fixture.atagged('gamma', 'delta'))
    self.assertEqual(2,            self.fixture.calls)
    self.assertEqual(4,            self.getSize())

    simpleKey = 'cache:entry:hermes.test:Fixture:asimple:' + self._arghash('alpha', 'beta')
    self.assertEqual(
      'ateb+ahpla', self.testee.mangler.loads(self.testee.backend.client.get(simpleKey))
    )

    rockTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:rock'))
    treeTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:tree'))
    self.assertFalse(rockTag == treeTag)
    self.assertEqual(22, len(rockTag))
    self.assertEqual(22, len(treeTag))

    argHash   = self._arghash('gamma', 'delta')
    tagHash   = self.testee.mangler.hashTags(dict(tree = treeTag, rock = rockTag))
    taggedKey = 'cache:entry:hermes.test:Fixture:atagged:{0}:{1}'.format(argHash, tagHash)
    self.assertEqual(
      'aldamg', self.testee.mangler.loads(self.testee.backend.client.get(taggedKey))
    )


    self.testee.clean(('rock',))
    self.assertEqual(3, self.getSize())

    self.assertEqual(
      'ateb+ahpla', self.testee.mangler.loads(self.testee.backend.client.get(simpleKey))
    )

    self.assertIsNone(self.testee.backend.client.get('cache:tag:rock'))
    self.assertEqual(
      treeTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:tree'))
    )

    # stale still accessible, though only directly
    self.assertEqual(
      'aldamg', self.testee.mangler.loads(self.testee.backend.client.get(taggedKey))
    )


    self.assertEqual('ateb+ahpla', await self.fixture.asimple('alpha', 'beta'))
    self.assertEqual('aldamg',     await self.fixture.atagged('gamma', 'delta'))
    self.assertEqual(3,            self.fixture.calls)
    self.assertEqual(5,            self.getSize(), '+1 old tagged entry')

    self.assertEqual(
      'ateb+ahpla', self.testee.mangler.loads(self.testee.backend.client.get(simpleKey))
    )

    self.assertNotEqual(
      rockTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:rock'))
    )
    rockTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:rock'))
    self.assertFalse(rockTag == treeTag)
    self.assertEqual(22, len(rockTag))
    self.assertEqual(
      treeTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:tree'))
    )

    # stale still accessible, though only directly
    self.assertEqual(
      'aldamg', self.testee.mangler.loads(self.testee.backend.client.get(taggedKey))
    )

    argHash   = self._arghash('gamma', 'delta')
    tagHash   = self.testee.mangler.hashTags(dict(tree = treeTag, rock = rockTag))
    taggedKey = 'cache:entry:hermes.test:Fixture:atagged:{0}:{1}'.format(argHash, tagHash)
    self.assertEqual(
      'aldamg', self.testee.mangler.loads(self.testee.backend.client.get(taggedKey))
    )


    self.testee.clean(('rock', 'tree'))
    self.assertEqual(3, self.getSize(), 'simaple, new tagged and old tagged')

    self.assertEqual(
      'ateb+ahpla', self.testee.mangler.loads(self.testee.backend.client.get(simpleKey))
    )

    # new stale is accessible, though only directly
    self.assertEqual(
      'aldamg', self.testee.mangler.loads(self.testee.backend.client.get(taggedKey))
    )


    self.assertEqual('ateb+ahpla', await self.fixture.asimple('alpha', 'beta'))
    self.assertEqual('aldamg',     await self.fixture.atagged('gamma', 'delta'))
    self.assertEqual(4,            self.fixture.calls)
    self.assertEqual(6,            self.getSize(), '+2 old tagged entries')

    self.assertEqual(
      'ateb+ahpla', self.testee.mangler.loads(self.testee.backend.client.get(simpleKey))
    )

    # new stale still accessible, though only directly
    self.assertEqual(
      'aldamg', self.testee.mangler.loads(self.testee.backend.client.get(taggedKey))
    )

    self.assertNotEqual(
      rockTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:rock'))
    )
    self.assertNotEqual(
      treeTag, self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:tree'))
    )

    rockTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:rock'))
    treeTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:tree'))
    self.assertFalse(rockTag == treeTag)
    self.assertEqual(22, len(rockTag))
    self.assertEqual(22, len(treeTag))

    argHash   = self._arghash('gamma', 'delta')
    tagHash   = self.testee.mangler.hashTags(dict(tree = treeTag, rock = rockTag))
    taggedKey = 'cache:entry:hermes.test:Fixture:atagged:{0}:{1}'.format(argHash, tagHash)
    self.assertEqual(
      'aldamg', self.testee.mangler.loads(self.testee.backend.client.get(taggedKey))
    )


    self.testee.clean(('rock', 'tree'))
    self.testee.clean()
    self.assertEmptyBackend()

  def testNested(self):
    self.assertEqual('beta+alpha', self.fixture.nested('alpha', 'beta'))
    self.assertEqual(2, self.fixture.calls)
    key = 'cache:entry:hermes.test:Fixture:nested:' + self._arghash('alpha', 'beta')
    self.assertEqual('beta+alpha', self.testee.mangler.loads(self.testee.backend.client.get(key)))
    key = 'cache:entry:hermes.test:Fixture:simple:' + self._arghash('beta', 'alpha')
    self.assertEqual('ahpla+ateb', self.testee.mangler.loads(self.testee.backend.client.get(key)))

  @asynctest
  async def testAsyncNested(self):
    self.assertEqual('beta+alpha', await self.fixture.anested('alpha', 'beta'))
    self.assertEqual(2, self.fixture.calls)
    key = 'cache:entry:hermes.test:Fixture:anested:' + self._arghash('alpha', 'beta')
    self.assertEqual('beta+alpha', self.testee.mangler.loads(self.testee.backend.client.get(key)))
    key = 'cache:entry:hermes.test:Fixture:asimple:' + self._arghash('beta', 'alpha')
    self.assertEqual('ahpla+ateb', self.testee.mangler.loads(self.testee.backend.client.get(key)))

  def testConcurrent(self):
    log = []
    key = lambda fn, *args, **kwargs: 'mk:{0}:{1}'.format(*args)

    @self.testee(tags = ('a', 'z'), key = key, ttl = 120)
    def bar(a, b):
      log.append(1)
      time.sleep(0.04)
      return '{0}-{1}'.format(a, b)[::2]

    threads = [threading.Thread(target = bar, args = ('alpha', 'beta')) for _ in range(4)]
    tuple(map(threading.Thread.start, threads))
    tuple(map(threading.Thread.join,  threads))

    self.assertEqual(1, sum(log))
    self.assertEqual(3, self.getSize())

    aTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:a'))
    zTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:z'))
    self.assertFalse(aTag == zTag)
    self.assertEqual(22, len(aTag))
    self.assertEqual(22, len(zTag))

    tagHash = self.testee.mangler.hashTags(dict(a = aTag, z = zTag))
    key     = 'mk:alpha:beta:' + tagHash
    self.assertEqual('apabt', self.testee.mangler.loads(self.testee.backend.client.get(key)))

    del log[:]
    self.testee.clean()
    self.testee.backend.lock = lambda k: backend.AbstractLock(k)  # now see a dogpile

    threads = [threading.Thread(target = bar, args = ('alpha', 'beta')) for _ in range(4)]
    tuple(map(threading.Thread.start, threads))
    tuple(map(threading.Thread.join,  threads))

    self.assertGreater(sum(log), 1, 'dogpile')
    # enries may be duplicated if tags overwrite
    self.assertGreaterEqual(self.getSize(), 3)

  @asynctest
  async def testAsyncConcurrent(self):
    log = []
    key = lambda fn, *args, **kwargs: 'mk:{0}:{1}'.format(*args)

    @self.testee(tags = ('a', 'z'), key = key, ttl = 120)
    async def bar(a, b):
      log.append(1)
      await asyncio.sleep(0.04)
      return '{0}-{1}'.format(a, b)[::2]

    await asyncio.gather(*(bar('alpha', 'beta') for _ in range(4)))

    self.assertEqual(1, sum(log))
    self.assertEqual(3, self.getSize())

    aTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:a'))
    zTag = self.testee.mangler.loads(self.testee.backend.client.get('cache:tag:z'))
    self.assertFalse(aTag == zTag)
    self.assertEqual(22, len(aTag))
    self.assertEqual(22, len(zTag))

    tagHash = self.testee.mangler.hashTags(dict(a = aTag, z = zTag))
    key     = 'mk:alpha:beta:' + tagHash
    self.assertEqual('apabt', self.testee.mangler.loads(self.testee.backend.client.get(key)))

    del log[:]
    self.testee.clean()
    self.testee.backend.lock = lambda k: backend.AbstractLock(k)  # now see a dogpile

    await asyncio.gather(*(bar('alpha', 'beta') for _ in range(4)))

    self.assertGreater(sum(log), 1, 'dogpile')
    # enries may be duplicated if tags overwrite
    self.assertGreaterEqual(self.getSize(), 3)


class BaseRemoteBackendLockTest(TestCase):

  ttlDelta = 0

  cache: Hermes
  testee: backend.AbstractLock

  def setUp(self):
    '''``cache` and ```testee`` are expected to be created in override.'''

    self.cache.clean()

  def getKeyTtl(self, key: str) -> int:
    raise NotImplementedError

  def testAcquire(self):
    for _ in range(2):
      try:
        self.assertTrue(self.testee.acquire(True))
        self.assertFalse(self.testee.acquire(False))
        self.assertEqual('123', self.testee.key)
        self.assertAlmostEqual(900, self.getKeyTtl(self.testee.key), delta = self.ttlDelta)
      finally:
        self.testee.release()

  def testRelease(self):
    for _ in range(2):
      try:
        self.assertTrue(self.testee.acquire(True))
        self.assertFalse(self.testee.acquire(False))
        self.assertEqual('123', self.testee.key)
      finally:
        self.testee.release()
      self.assertIs(None, self.testee.client.get(self.testee.key))

  def testWith(self):
    with self.testee:
      self.assertFalse(self.testee.acquire(False))
      self.assertEqual('123', self.testee.key)

      another = type(self.testee)('234', self.cache.backend.client)
      with another:
        self.assertFalse(another.acquire(False))
        self.assertFalse(self.testee.acquire(False))
        self.assertEqual('234', another.key)

  def testConcurrent(self):
    log = []
    check = threading.Lock()

    def target():
      with self.testee:
        log.append(check.acquire(False))
        time.sleep(0.05)
        check.release()
        time.sleep(0.05)

    threads = tuple(map(lambda _: threading.Thread(target = target), range(4)))
    tuple(map(threading.Thread.start, threads))
    tuple(map(threading.Thread.join,  threads))

    self.assertEqual([True] * 4, log)

  @asynctest
  async def testAsyncConcurrent(self):
    log = []
    check = threading.Lock()

    loop = asyncio.get_event_loop()

    async def target():
      await loop.run_in_executor(None, self.testee.acquire)
      try:
        log.append(check.acquire(False))
        await asyncio.sleep(0.05)
        check.release()
        await asyncio.sleep(0.05)
      finally:
        await loop.run_in_executor(None, self.testee.release)

    await asyncio.gather(*(target() for _ in range(4)))

    self.assertEqual([True] * 4, log)



from typing import TYPE_CHECKING, Any, Dict, Iterable, Optional, Union


if TYPE_CHECKING:
  from .. import Mangler  # @UnusedImport


__all__ = 'AbstractBackend',


class AbstractLock:
  '''
  Base locking class. Implements context manger protocol. Mocks
  ``acquire`` and ``release`` i.e. it always acquires.
  '''

  key: str
  '''Implementation may be key-aware.'''

  def __init__(self, key: str):
    self.key = key

  def __enter__(self):
    '''Enter context manager by acquiring the distributed lock.'''

    self.acquire()

  def __exit__(self, type, value, traceback):
    '''Exit context manager by releasing the distributed lock.'''

    self.release()

  def acquire(self, wait = True) -> bool:
    '''
    Acquire distributed lock.

    :argument wait: Whether to wait for the lock.
    :return: Whether the lock was acquired.
    '''

    return True

  def release(self):
    '''Release distributed lock.'''


class AbstractBackend:
  '''Base backend class. It's also the no-op implementation.'''

  mangler: 'Mangler'
  '''Key manager responsible for creating keys, hashing and serialisation.'''

  def __init__(self, mangler: 'Mangler'):
    self.mangler = mangler

  def lock(self, key: str) -> AbstractLock:
    '''Create lock object for the key.'''

    return AbstractLock(self.mangler.nameLock(key))

  def save(self, mapping: Dict[str, Any], *, ttl: Optional[int] = None):
    '''
    Save cache entry.

    :argument mapping: ``key``-``value`` mapping for a bulk save.
    :argument ttl: Cache entry time-to-live .
    '''

  def load(self, keys: Union[str, Iterable[str]]) -> Optional[Union[Any, Dict[str, Any]]]:
    '''
    Load cache entry(ies).

    Note, when handling a multiple key call, absent value keys
    should be excluded from resulting dictionary.
    '''

    return None if isinstance(keys, str) else {}

  def remove(self, keys: Union[str, Iterable[str]]):
    '''Remove given keys.'''

  def clean(self):
    '''Purge the backend storage.'''



import asyncio
import gc
import platform
import threading
import time

from .. import Hermes, test
from ..backend import AbstractLock, inprocess
from . import asynctest


class TestInprocess(test.TestCase):

  def setUp(self):
    self.testee = Hermes(inprocess.Backend, ttl = 360, ttlWatchSleep = 0.01)
    self.addCleanup(self.testee.backend.stopWatch)

    self.fixture = test.createFixture(self.testee)

    self.testee.clean()

  def testSimple(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

    for _ in range(4):
      self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', b = 'beta'))
      self.assertEqual(1, self.fixture.calls)

      key = 'cache:entry:hermes.test:Fixture:simple:' + self._arghash('alpha', b = 'beta')
      self.assertEqual({key: 'ateb+ahpla'}, self.testee.backend.dump())

    self.fixture.simple.invalidate('alpha', b = 'beta')
    self.assertEqual({}, self.testee.backend.dump())


    self.assertEqual(1,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

    expected = "]}'ammag'{[+}]'ateb'[ :'ahpla'{"
    for _ in range(4):
      self.assertEqual(expected, self.fixture.simple({'alpha': ['beta']}, [{'gamma'}]))

      self.assertEqual(2, self.fixture.calls)
      key = 'cache:entry:hermes.test:Fixture:simple:{}'.format(
        self._arghash({'alpha': ['beta']}, [{'gamma'}]))
      self.assertEqual({key: expected}, self.testee.backend.dump())


    self.fixture.simple.invalidate({'alpha': ['beta']}, [{'gamma'}])
    self.assertEqual({}, self.testee.backend.dump())

  def testTagged(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

    for _ in range(4):
      self.assertEqual('ae-hl', self.fixture.tagged('alpha', b = 'beta'))
      self.assertEqual(1,       self.fixture.calls)

      cache = self.testee.backend.dump()
      self.assertEqual(3, len(cache))
      self.assertEqual(22, len(cache.pop('cache:tag:rock')))
      self.assertEqual(22, len(cache.pop('cache:tag:tree')))

      expected = 'cache:entry:hermes.test:Fixture:tagged:' + self._arghash('alpha', b = 'beta')
      self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
      self.assertEqual('ae-hl', tuple(cache.values())[0])


    self.fixture.tagged.invalidate('alpha', b = 'beta')

    cache = self.testee.backend.dump()
    self.assertEqual(2, len(cache))
    rockTag = cache.get('cache:tag:rock')
    treeTag = cache.get('cache:tag:tree')
    self.assertNotEqual(rockTag, treeTag)
    self.assertEqual(22, len(rockTag))
    self.assertEqual(22, len(treeTag))

    for _ in range(4):
      self.assertEqual('ae-hl', self.fixture.tagged('alpha', 'beta'))
      self.assertEqual('ae%hl', self.fixture.tagged2('alpha', 'beta'))
      self.assertEqual(3, self.fixture.calls)

      cache = self.testee.backend.dump()
      self.assertEqual(5, len(cache))
      self.assertEqual(rockTag, cache.get('cache:tag:rock'))
      self.assertEqual(treeTag, cache.get('cache:tag:tree'))
      self.assertEqual(22, len(cache.get('cache:tag:ice')))

    self.testee.clean(['rock'])

    cache = self.testee.backend.dump()
    self.assertEqual(4, len(cache))
    self.assertTrue('cache:tag:rock' not in cache)
    iceTag = cache.get('cache:tag:ice')

    for _ in range(4):
      self.assertEqual('ae-hl', self.fixture.tagged('alpha', 'beta'))
      self.assertEqual('ae%hl', self.fixture.tagged2('alpha', 'beta'))
      self.assertEqual(5, self.fixture.calls)

      cache = self.testee.backend.dump()
      self.assertEqual(7,  len(cache), 'has new and old entries for tagged and tagged 2 + 3 tags')
      self.assertEqual(treeTag, cache.get('cache:tag:tree'))
      self.assertEqual(iceTag,  cache.get('cache:tag:ice'))
      self.assertEqual(22, len(cache.get('cache:tag:rock')))
      self.assertNotEqual(rockTag, cache.get('cache:tag:rock'))

  def testFunction(self):
    counter = dict(foo = 0, bar = 0)

    @self.testee
    def foo(a, b):
      counter['foo'] += 1
      return '{0}+{1}'.format(a, b)[::-1]

    key = lambda _fn, *args, **kwargs: 'mk:{0}:{1}'.format(*args)

    @self.testee(tags = ('a', 'z'), key = key, ttl = 120)
    def bar(a, b):
      counter['bar'] += 1
      return '{0}-{1}'.format(a, b)[::2]


    self.assertEqual(0,  counter['foo'])
    self.assertEqual({}, self.testee.backend.dump())

    for _ in range(4):
      self.assertEqual('ateb+ahpla', foo('alpha', 'beta'))
      self.assertEqual(1, counter['foo'])
      self.assertEqual(
        {'cache:entry:hermes.test.inprocess:foo:' + self._arghash('alpha', 'beta'): 'ateb+ahpla'},
        self.testee.backend.dump()
      )

    foo.invalidate('alpha', 'beta')
    self.assertEqual(1,  counter['foo'])
    self.assertEqual({}, self.testee.backend.dump())


    self.assertEqual(0,  counter['bar'])
    self.assertEqual({}, self.testee.backend.dump())

    for _ in range(4):
      self.assertEqual('apabt', bar('alpha', 'beta'))
      self.assertEqual(1,       counter['bar'])

      cache = self.testee.backend.dump()
      self.assertEqual(3,  len(cache))
      self.assertEqual(22, len(cache.pop('cache:tag:a')))
      self.assertEqual(22, len(cache.pop('cache:tag:z')))

      expected = 'mk:alpha:beta'
      self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
      self.assertEqual('apabt', tuple(cache.values())[0])

    bar.invalidate('alpha', 'beta')
    self.assertEqual(1, counter['foo'])

    cache = self.testee.backend.dump()
    self.assertEqual(2,  len(cache))
    self.assertEqual(22, len(cache.pop('cache:tag:a')))
    self.assertEqual(22, len(cache.pop('cache:tag:z')))

  def testKey(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

    for _ in range(4):
      self.assertEqual('apabt', self.fixture.key('alpha', 'beta'))
      self.assertEqual(1,       self.fixture.calls)

      cache = self.testee.backend.dump()
      self.assertEqual(3,  len(cache))
      self.assertEqual(22, len(cache.pop('cache:tag:ash')))
      self.assertEqual(22, len(cache.pop('cache:tag:stone')))

      expected = 'mykey:alpha:beta'
      self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
      self.assertEqual('apabt', tuple(cache.values())[0])

    self.fixture.key.invalidate('alpha', 'beta')

    cache = self.testee.backend.dump()
    self.assertEqual(2,  len(cache))
    self.assertEqual(22, len(cache.pop('cache:tag:ash')))
    self.assertEqual(22, len(cache.pop('cache:tag:stone')))

  def testAll(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

    for _ in range(4):
      self.assertEqual({'a': 1, 'b': {'b': 'beta'}}, self.fixture.all({'alpha': 1}, ['beta']))
      self.assertEqual(1, self.fixture.calls)

      cache = self.testee.backend.dump()
      self.assertEqual(3, len(cache))
      self.assertEqual(22, len(cache.pop('cache:tag:a')))
      self.assertEqual(22, len(cache.pop('cache:tag:z')))

      expected = "mk:{'alpha':1}:['beta']"
      self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
      self.assertEqual({'a': 1, 'b': {'b': 'beta'}}, tuple(cache.values())[0])

    self.fixture.all.invalidate({'alpha': 1}, ['beta'])

    cache = self.testee.backend.dump()
    self.assertEqual(2,  len(cache))
    self.assertEqual(22, len(cache.pop('cache:tag:a')))
    self.assertEqual(22, len(cache.pop('cache:tag:z')))

  def testClean(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(2,            self.fixture.calls)

    self.testee.clean()

    self.assertEqual(2,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

  def testCleanTagged(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(2,            self.fixture.calls)

    cache = self.testee.backend.dump()
    self.assertEqual(4, len(cache))

    self.assertEqual(
      'ateb+ahpla',
      cache.pop(
        'cache:entry:hermes.test:Fixture:simple:' + self._arghash('alpha', 'beta')
      ),
    )
    self.assertEqual(22, len(cache.pop('cache:tag:rock')))
    self.assertEqual(22, len(cache.pop('cache:tag:tree')))

    expectedKey = 'cache:entry:hermes.test:Fixture:tagged:' + self._arghash('gamma', 'delta')
    self.assertEqual(expectedKey, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('aldamg', tuple(cache.values())[0])


    self.testee.clean(('rock',))

    cache = self.testee.backend.dump()
    self.assertEqual(3, len(cache))
    self.assertEqual(
      'ateb+ahpla',
      cache.pop('cache:entry:hermes.test:Fixture:simple:' + self._arghash('alpha', 'beta')))
    self.assertEqual(22, len(cache.pop('cache:tag:tree')))
    self.assertFalse('cache:tag:rock' in cache)

    expectedKey = 'cache:entry:hermes.test:Fixture:tagged:' + self._arghash('gamma', 'delta')
    self.assertEqual(expectedKey, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('aldamg', tuple(cache.values())[0])


    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(3,            self.fixture.calls)

    cache = self.testee.backend.dump()
    self.assertEqual(5, len(cache), '+1 old tagged entry')
    self.assertEqual(
      'ateb+ahpla',
      cache.pop('cache:entry:hermes.test:Fixture:simple:' + self._arghash('alpha', 'beta')))
    self.assertEqual(22, len(cache.pop('cache:tag:rock')))
    self.assertEqual(22, len(cache.pop('cache:tag:tree')))

    expectedKey = 'cache:entry:hermes.test:Fixture:tagged:' + self._arghash('gamma', 'delta')
    self.assertEqual(expectedKey, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('aldamg', tuple(cache.values())[0])


    self.testee.clean(('rock', 'tree'))

    cache = self.testee.backend.dump()
    self.assertEqual(3, len(cache))
    self.assertEqual(
      'ateb+ahpla',
      cache.pop('cache:entry:hermes.test:Fixture:simple:' + self._arghash('alpha', 'beta')))
    self.assertFalse('cache:tag:tree' in cache)
    self.assertFalse('cache:tag:rock' in cache)

    expectedKey = 'cache:entry:hermes.test:Fixture:tagged:' + self._arghash('gamma', 'delta')
    self.assertEqual(expectedKey, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('aldamg', tuple(cache.values())[0])


    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(4,            self.fixture.calls)

    cache = self.testee.backend.dump()
    self.assertEqual(6, len(cache), '+2 old tagged entries')
    self.assertEqual(
      'ateb+ahpla',
      cache.pop(
        'cache:entry:hermes.test:Fixture:simple:' + self._arghash('alpha', 'beta')
      ),
    )
    expectedKey = 'cache:entry:hermes.test:Fixture:tagged:' + self._arghash('gamma', 'delta')
    self.assertTrue(expectedKey in (':'.join(k.split(':')[:-1]) for k in cache.keys()))
    self.assertTrue('aldamg' in cache.values())


    self.testee.clean()

    self.assertEqual(4,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

  def testNested(self):
    self.assertEqual('beta+alpha', self.fixture.nested('alpha', 'beta'))
    self.assertEqual(2, self.fixture.calls)
    self.assertEqual({
      'cache:entry:hermes.test:Fixture:nested:' + self._arghash('alpha', 'beta'): 'beta+alpha',
      'cache:entry:hermes.test:Fixture:simple:' + self._arghash('beta', 'alpha'): 'ahpla+ateb'
    }, self.testee.backend.dump())

  def testConcurrent(self):
    log = []
    key = lambda _fn, *args, **kwargs: 'mk:{0}:{1}'.format(*args)

    @self.testee(tags = ('a', 'z'), key = key, ttl = 120)
    def bar(a, b):
      log.append(1)
      time.sleep(0.04)
      return '{0}-{1}'.format(a, b)[::2]

    threads = [threading.Thread(target = bar, args = ('alpha', 'beta')) for _ in range(4)]
    tuple(map(threading.Thread.start, threads))
    tuple(map(threading.Thread.join,  threads))

    self.assertEqual(1, sum(log))

    cache = self.testee.backend.dump()
    self.assertEqual(22, len(cache.pop('cache:tag:a')))
    self.assertEqual(22, len(cache.pop('cache:tag:z')))

    self.assertEqual('mk:alpha:beta', ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('apabt', tuple(cache.values())[0])

    del log[:]
    self.testee.clean()
    self.testee.backend.lock = lambda k: AbstractLock(k)  # now see a dogpile

    threads = [threading.Thread(target = bar, args = ('alpha', 'beta')) for _ in range(4)]
    tuple(map(threading.Thread.start, threads))
    tuple(map(threading.Thread.join,  threads))

    self.assertGreater(sum(log), 1, 'dogpile')

    cache = self.testee.backend.dump()
    self.assertEqual(22, len(cache.pop('cache:tag:a')))
    self.assertEqual(22, len(cache.pop('cache:tag:z')))

    self.assertEqual('mk:alpha:beta', ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('apabt', tuple(cache.values())[0])

  def testExpiration(self):
    self.assertEqual({}, self.testee.backend.dump())

    @self.testee(ttl = 0.05)
    def fn1(a, b):
      return '{0}-{1}'.format(a, b)[::2]

    @self.testee(ttl = 0.2)
    def fn2(a, b):
      return '{0}-{1}'.format(a, b)[::2]

    fn1('alpha', b = 'beta')
    fn2('alpha', b = 'beta')

    key1 = 'cache:entry:hermes.test.inprocess:fn1:' + self._arghash('alpha', b = 'beta')
    key2 = 'cache:entry:hermes.test.inprocess:fn2:' + self._arghash('alpha', b = 'beta')
    self.assertIn(key1, self.testee.backend.dump())
    self.assertIn(key2, self.testee.backend.dump())

    time.sleep(0.1)
    self.assertNotIn(key1, self.testee.backend.dump())
    self.assertIn(key2, self.testee.backend.dump())


class TestAsyncInprocess(test.TestCase):

  def setUp(self):
    self.testee = Hermes(inprocess.AsyncBackend, ttl = 360, ttlWatchSleep = 0.01)
    self.fixture = test.createFixture(self.testee)

    self.testee.clean()

  def tearDown(self):
    super().tearDown()

    if platform.python_implementation() == 'PyPy':
      gc.collect()

    self.assertEqual([], self.testee.backend._lockMap.valuerefs())

  @asynctest
  async def testAsyncSimple(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

    for _ in range(4):
      self.assertEqual('ateb+ahpla', await self.fixture.asimple('alpha', b = 'beta'))
      self.assertEqual(1, self.fixture.calls)

      key = 'cache:entry:hermes.test:Fixture:asimple:' + self._arghash('alpha', b = 'beta')
      self.assertEqual({key: 'ateb+ahpla'}, self.testee.backend.dump())

    await self.fixture.asimple.invalidate('alpha', b = 'beta')
    self.assertEqual({}, self.testee.backend.dump())


    self.assertEqual(1,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

    expected = "]}'ammag'{[+}]'ateb'[ :'ahpla'{"
    for _ in range(4):
      self.assertEqual(expected, await self.fixture.asimple({'alpha': ['beta']}, [{'gamma'}]))

      self.assertEqual(2, self.fixture.calls)
      key = 'cache:entry:hermes.test:Fixture:asimple:{}'.format(
        self._arghash({'alpha': ['beta']}, [{'gamma'}]))
      self.assertEqual({key: expected}, self.testee.backend.dump())


    await self.fixture.asimple.invalidate({'alpha': ['beta']}, [{'gamma'}])
    self.assertEqual({}, self.testee.backend.dump())

  @asynctest
  async def testAsyncTagged(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

    for _ in range(4):
      self.assertEqual('ae-hl', await self.fixture.atagged('alpha', b = 'beta'))
      self.assertEqual(1,       self.fixture.calls)

      cache = self.testee.backend.dump()
      self.assertEqual(3, len(cache))
      self.assertEqual(22, len(cache.pop('cache:tag:rock')))
      self.assertEqual(22, len(cache.pop('cache:tag:tree')))

      expected = 'cache:entry:hermes.test:Fixture:atagged:' + self._arghash('alpha', b = 'beta')
      self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
      self.assertEqual('ae-hl', tuple(cache.values())[0])


    await self.fixture.atagged.invalidate('alpha', b = 'beta')

    cache = self.testee.backend.dump()
    self.assertEqual(2, len(cache))
    rockTag = cache.get('cache:tag:rock')
    treeTag = cache.get('cache:tag:tree')
    self.assertNotEqual(rockTag, treeTag)
    self.assertEqual(22, len(rockTag))
    self.assertEqual(22, len(treeTag))

    for _ in range(4):
      self.assertEqual('ae-hl', await self.fixture.atagged('alpha', 'beta'))
      self.assertEqual('ae%hl', await self.fixture.atagged2('alpha', 'beta'))
      self.assertEqual(3, self.fixture.calls)

      cache = self.testee.backend.dump()
      self.assertEqual(5, len(cache))
      self.assertEqual(rockTag, cache.get('cache:tag:rock'))
      self.assertEqual(treeTag, cache.get('cache:tag:tree'))
      self.assertEqual(22, len(cache.get('cache:tag:ice')))

    self.testee.clean(['rock'])

    cache = self.testee.backend.dump()
    self.assertEqual(4, len(cache))
    self.assertTrue('cache:tag:rock' not in cache)
    iceTag = cache.get('cache:tag:ice')

    for _ in range(4):
      self.assertEqual('ae-hl', await self.fixture.atagged('alpha', 'beta'))
      self.assertEqual('ae%hl', await self.fixture.atagged2('alpha', 'beta'))
      self.assertEqual(5, self.fixture.calls)

      cache = self.testee.backend.dump()
      self.assertEqual(7,  len(cache), 'has new and old entries for tagged and tagged 2 + 3 tags')
      self.assertEqual(treeTag, cache.get('cache:tag:tree'))
      self.assertEqual(iceTag,  cache.get('cache:tag:ice'))
      self.assertEqual(22, len(cache.get('cache:tag:rock')))
      self.assertNotEqual(rockTag, cache.get('cache:tag:rock'))

  @asynctest
  async def testAsyncFunction(self):
    counter = dict(foo = 0, bar = 0)

    @self.testee
    async def foo(a, b):
      await asyncio.sleep(0)
      counter['foo'] += 1
      return '{0}+{1}'.format(a, b)[::-1]

    key = lambda _fn, *args, **kwargs: 'mk:{0}:{1}'.format(*args)

    @self.testee(tags = ('a', 'z'), key = key, ttl = 120)
    async def bar(a, b):
      await asyncio.sleep(0)
      counter['bar'] += 1
      return '{0}-{1}'.format(a, b)[::2]


    self.assertEqual(0,  counter['foo'])
    self.assertEqual({}, self.testee.backend.dump())

    for _ in range(4):
      self.assertEqual('ateb+ahpla', await foo('alpha', 'beta'))
      self.assertEqual(1, counter['foo'])
      self.assertEqual(
        {'cache:entry:hermes.test.inprocess:foo:' + self._arghash('alpha', 'beta'): 'ateb+ahpla'},
        self.testee.backend.dump()
      )

    await foo.invalidate('alpha', 'beta')
    self.assertEqual(1,  counter['foo'])
    self.assertEqual({}, self.testee.backend.dump())


    self.assertEqual(0,  counter['bar'])
    self.assertEqual({}, self.testee.backend.dump())

    for _ in range(4):
      self.assertEqual('apabt', await bar('alpha', 'beta'))
      self.assertEqual(1,       counter['bar'])

      cache = self.testee.backend.dump()
      self.assertEqual(3,  len(cache))
      self.assertEqual(22, len(cache.pop('cache:tag:a')))
      self.assertEqual(22, len(cache.pop('cache:tag:z')))

      expected = 'mk:alpha:beta'
      self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
      self.assertEqual('apabt', tuple(cache.values())[0])

    await bar.invalidate('alpha', 'beta')
    self.assertEqual(1, counter['foo'])

    cache = self.testee.backend.dump()
    self.assertEqual(2,  len(cache))
    self.assertEqual(22, len(cache.pop('cache:tag:a')))
    self.assertEqual(22, len(cache.pop('cache:tag:z')))

  @asynctest
  async def testAsyncKey(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

    for _ in range(4):
      self.assertEqual('apabt', await self.fixture.akey('alpha', 'beta'))
      self.assertEqual(1,       self.fixture.calls)

      cache = self.testee.backend.dump()
      self.assertEqual(3,  len(cache))
      self.assertEqual(22, len(cache.pop('cache:tag:ash')))
      self.assertEqual(22, len(cache.pop('cache:tag:stone')))

      expected = 'mykey:alpha:beta'
      self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
      self.assertEqual('apabt', tuple(cache.values())[0])

    await self.fixture.akey.invalidate('alpha', 'beta')

    cache = self.testee.backend.dump()
    self.assertEqual(2,  len(cache))
    self.assertEqual(22, len(cache.pop('cache:tag:ash')))
    self.assertEqual(22, len(cache.pop('cache:tag:stone')))

  @asynctest
  async def testAsyncAll(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

    for _ in range(4):
      self.assertEqual(
        {'a': 1, 'b': {'b': 'beta'}}, await self.fixture.aall({'alpha': 1}, ['beta'])
      )
      self.assertEqual(1, self.fixture.calls)

      cache = self.testee.backend.dump()
      self.assertEqual(3, len(cache))
      self.assertEqual(22, len(cache.pop('cache:tag:a')))
      self.assertEqual(22, len(cache.pop('cache:tag:z')))

      expected = "mk:{'alpha':1}:['beta']"
      self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
      self.assertEqual({'a': 1, 'b': {'b': 'beta'}}, tuple(cache.values())[0])

    await self.fixture.aall.invalidate({'alpha': 1}, ['beta'])

    cache = self.testee.backend.dump()
    self.assertEqual(2,  len(cache))
    self.assertEqual(22, len(cache.pop('cache:tag:a')))
    self.assertEqual(22, len(cache.pop('cache:tag:z')))

  @asynctest
  async def testAsyncClean(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

    self.assertEqual('ateb+ahpla', await self.fixture.asimple('alpha', 'beta'))
    self.assertEqual('aldamg',     await self.fixture.atagged('gamma', 'delta'))
    self.assertEqual(2,            self.fixture.calls)

    self.testee.clean()

    self.assertEqual(2,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

  @asynctest
  async def testAsyncCleanTagged(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

    self.assertEqual('ateb+ahpla', await self.fixture.asimple('alpha', 'beta'))
    self.assertEqual('aldamg',     await self.fixture.atagged('gamma', 'delta'))
    self.assertEqual(2,            self.fixture.calls)

    cache = self.testee.backend.dump()
    self.assertEqual(4, len(cache))

    self.assertEqual(
      'ateb+ahpla',
      cache.pop(
        'cache:entry:hermes.test:Fixture:asimple:' + self._arghash('alpha', 'beta')
      ),
    )
    self.assertEqual(22, len(cache.pop('cache:tag:rock')))
    self.assertEqual(22, len(cache.pop('cache:tag:tree')))

    expectedKey = 'cache:entry:hermes.test:Fixture:atagged:' + self._arghash('gamma', 'delta')
    self.assertEqual(expectedKey, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('aldamg', tuple(cache.values())[0])


    self.testee.clean(('rock',))

    cache = self.testee.backend.dump()
    self.assertEqual(3, len(cache))
    self.assertEqual(
      'ateb+ahpla',
      cache.pop('cache:entry:hermes.test:Fixture:asimple:' + self._arghash('alpha', 'beta')))
    self.assertEqual(22, len(cache.pop('cache:tag:tree')))
    self.assertFalse('cache:tag:rock' in cache)

    expectedKey = 'cache:entry:hermes.test:Fixture:atagged:' + self._arghash('gamma', 'delta')
    self.assertEqual(expectedKey, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('aldamg', tuple(cache.values())[0])


    self.assertEqual('ateb+ahpla', await self.fixture.asimple('alpha', 'beta'))
    self.assertEqual('aldamg',     await self.fixture.atagged('gamma', 'delta'))
    self.assertEqual(3,            self.fixture.calls)

    cache = self.testee.backend.dump()
    self.assertEqual(5, len(cache), '+1 old tagged entry')
    self.assertEqual(
      'ateb+ahpla',
      cache.pop('cache:entry:hermes.test:Fixture:asimple:' + self._arghash('alpha', 'beta')))
    self.assertEqual(22, len(cache.pop('cache:tag:rock')))
    self.assertEqual(22, len(cache.pop('cache:tag:tree')))

    expectedKey = 'cache:entry:hermes.test:Fixture:atagged:' + self._arghash('gamma', 'delta')
    self.assertEqual(expectedKey, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('aldamg', tuple(cache.values())[0])


    self.testee.clean(('rock', 'tree'))

    cache = self.testee.backend.dump()
    self.assertEqual(3, len(cache))
    self.assertEqual(
      'ateb+ahpla',
      cache.pop('cache:entry:hermes.test:Fixture:asimple:' + self._arghash('alpha', 'beta')))
    self.assertFalse('cache:tag:tree' in cache)
    self.assertFalse('cache:tag:rock' in cache)

    expectedKey = 'cache:entry:hermes.test:Fixture:atagged:' + self._arghash('gamma', 'delta')
    self.assertEqual(expectedKey, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('aldamg', tuple(cache.values())[0])


    self.assertEqual('ateb+ahpla', await self.fixture.asimple('alpha', 'beta'))
    self.assertEqual('aldamg',     await self.fixture.atagged('gamma', 'delta'))
    self.assertEqual(4,            self.fixture.calls)

    cache = self.testee.backend.dump()
    self.assertEqual(6, len(cache), '+2 old tagged entries')
    self.assertEqual(
      'ateb+ahpla',
      cache.pop(
        'cache:entry:hermes.test:Fixture:asimple:' + self._arghash('alpha', 'beta')
      ),
    )
    expectedKey = 'cache:entry:hermes.test:Fixture:atagged:' + self._arghash('gamma', 'delta')
    self.assertTrue(expectedKey in (':'.join(k.split(':')[:-1]) for k in cache.keys()))
    self.assertTrue('aldamg' in cache.values())


    self.testee.clean()

    self.assertEqual(4,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

  @asynctest
  async def testAsyncNested(self):
    self.assertEqual('beta+alpha', await self.fixture.anested('alpha', 'beta'))
    self.assertEqual(2, self.fixture.calls)
    self.assertEqual({
      'cache:entry:hermes.test:Fixture:anested:' + self._arghash('alpha', 'beta'): 'beta+alpha',
      'cache:entry:hermes.test:Fixture:asimple:' + self._arghash('beta', 'alpha'): 'ahpla+ateb'
    }, self.testee.backend.dump())

  @asynctest
  async def testAsyncConcurrent(self):
    log = []
    key = lambda _fn, *args, **kwargs: 'mk:{0}:{1}'.format(*args)

    @self.testee(tags = ('a', 'z'), key = key, ttl = 120)
    async def bar(a, b):
      log.append(1)
      await asyncio.sleep(0.04)
      return '{0}-{1}'.format(a, b)[::2]

    await asyncio.gather(*(bar('alpha', 'beta') for _ in range(4)))

    self.assertEqual(1, sum(log))

    cache = self.testee.backend.dump()
    self.assertEqual(22, len(cache.pop('cache:tag:a')))
    self.assertEqual(22, len(cache.pop('cache:tag:z')))

    self.assertEqual('mk:alpha:beta', ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('apabt', tuple(cache.values())[0])

    del log[:]
    self.testee.clean()
    self.testee.backend.lock = lambda k: AbstractLock(k)  # now see a dogpile

    await asyncio.gather(*(bar('alpha', 'beta') for _ in range(4)))

    self.assertGreater(sum(log), 1, 'dogpile')

    cache = self.testee.backend.dump()
    self.assertEqual(22, len(cache.pop('cache:tag:a')))
    self.assertEqual(22, len(cache.pop('cache:tag:z')))

    self.assertEqual('mk:alpha:beta', ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('apabt', tuple(cache.values())[0])

  @asynctest
  async def testAsyncExpiration(self):
    self.addCleanup(self.testee.backend.stopWatch)
    self.testee.backend.startWatch()

    self.assertEqual({}, self.testee.backend.dump())

    @self.testee(ttl = 0.05)
    async def fn1(a, b):
      await asyncio.sleep(0)
      return '{0}-{1}'.format(a, b)[::2]

    @self.testee(ttl = 0.2)
    def fn2(a, b):
      return '{0}-{1}'.format(a, b)[::2]

    await fn1('alpha', b = 'beta')
    fn2('alpha', b = 'beta')

    key1 = 'cache:entry:hermes.test.inprocess:fn1:' + self._arghash('alpha', b = 'beta')
    key2 = 'cache:entry:hermes.test.inprocess:fn2:' + self._arghash('alpha', b = 'beta')
    self.assertIn(key1, self.testee.backend.dump())
    self.assertIn(key2, self.testee.backend.dump())

    await asyncio.sleep(0.1)
    self.assertNotIn(key1, self.testee.backend.dump())
    self.assertIn(key2, self.testee.backend.dump())


class TestInprocessLock(test.TestCase):

  def setUp(self):
    self.testee = inprocess.Lock()

  def testAcquire(self):
    for _ in range(2):
      try:
        self.assertTrue(self.testee.acquire(True))
        self.assertTrue(self.testee.acquire(False))  # reentrant within one thread
      finally:
        self.testee.release()

  def testRelease(self):
    for _ in range(2):
      try:
        self.assertTrue(self.testee.acquire(True))
        self.assertTrue(self.testee.acquire(False))  # reentrant within one thread
      finally:
        self.testee.release()

  def testWith(self):
    with self.testee:
      self.assertTrue(self.testee.acquire(False))  # reentrant within one thread

  def testConcurrent(self):
    log   = []
    check = threading.Lock()

    def target():
      with self.testee:
        log.append(check.acquire(False))
        time.sleep(0.05)
        check.release()
        time.sleep(0.05)

    threads = [threading.Thread(target = target) for _ in range(4)]
    tuple(map(threading.Thread.start, threads))
    tuple(map(threading.Thread.join,  threads))

    self.assertEqual([True] * 4, log)


class TestAsyncInprocessLock(test.TestCase):

  def setUp(self):
    self.testee = inprocess.AsyncLock(123)

  @asynctest
  async def testAcquire(self):
    for _ in range(2):
      try:
        self.assertTrue(await self.testee.acquire(True))
        self.assertFalse(await self.testee.acquire(False))
      finally:
        await self.testee.release()

  @asynctest
  async def testRelease(self):
    for _ in range(2):
      try:
        self.assertTrue(await self.testee.acquire(True))
        self.assertFalse(await self.testee.acquire(False))
      finally:
        await self.testee.release()

  def testWith(self):
    # context manager protocol is no-op because synchronous code in in
    # asynchronous application executes exclusively be definition.
    # CachedCoro explicitly doesn't use the context manager but only
    # acquire() and release() directly.
    with self.testee:
      pass

  @asynctest
  async def testConcurrent(self):
    log = []
    check = threading.Lock()

    async def target():
      await self.testee.acquire()
      try:
        log.append(check.acquire(False))
        await asyncio.sleep(0.05)
        check.release()
        await asyncio.sleep(0.05)
      finally:
        await self.testee.release()

    await asyncio.gather(*(target() for _ in range(4)))

    self.assertEqual([True] * 4, log)



import builtins
import inspect
import json
import lzma
import os
import threading
import time
import types
import unittest
import zlib
from unittest import mock

from .. import Cached, CachedCoro, Compressor, Hermes, HermesError, Mangler, Serialiser, test
from ..backend import AbstractBackend, AbstractLock, inprocess
from . import asynctest


@mock.patch.object(builtins, 'print', mock.MagicMock())
class TestManualSnippets(unittest.TestCase):

  def testQuickstart(self):
    import hermes.backend.redis


    cache = hermes.Hermes(
      hermes.backend.redis.Backend,
      host = os.getenv('TEST_REDIS_HOST', 'localhost'),
      ttl = 600,
      db = 1,
    )

    @cache
    def foo(a, b):
      return a * b

    class Example:

      @cache(tags = ('math', 'power'), ttl = 1200)
      def bar(self, a, b):
        return a ** b

      @cache(tags = ('math', 'avg'), key = lambda fn, a, b: f'avg:{a}:{b}')
      def baz(self, a, b):
        return (a + b) / 2.0

    print(foo(2, 333))

    example = Example()
    print(example.bar(2, 10))
    print(example.baz(2, 10))

    foo.invalidate(2, 333)
    example.bar.invalidate(2, 10)
    example.baz.invalidate(2, 10)

    cache.clean(['math'])  # invalidate entries tagged 'math'
    cache.clean()          # flush cache

  def testNonTagged(self):
    import hermes.backend.inprocess  # noqa


    cache = hermes.Hermes(hermes.backend.inprocess.Backend)

    @cache
    def foo(a, b):
      return a * b

    foo(2, 2)
    foo(2, 4)

    print(cache.backend.dump())
    #  {
    #    'cache:entry:foo:bNapzUm_P8fjh8lhIWYgkg': 8,
    #    'cache:entry:foo:JjwE0zQhMRt_5lfwNNPk1Q': 4
    #  }

  def testTagged(self):
    import hermes.backend.inprocess  # noqa


    cache = hermes.Hermes(hermes.backend.inprocess.Backend)

    @cache(tags = ('tag1', 'tag2'))
    def foo(a, b):
      return a * b

    foo(2, 2)

    print(cache.backend.dump())
    #  {
    #    'cache:tag:tag1': 'SeL9JQMnT-_4jRN1nxeoeg',
    #    'cache:tag:tag2': 'n_OXAKlSjiz5zK5iAEe7zA',
    #    'cache:entry:foo:JjwE0zQhMRt_5lfwNNPk1Q:h_C-2mB3kWNNrA4r7byxNA': 4
    #  }

  def testTradeoff(self):
    import hermes.backend.inprocess  # noqa


    cache = hermes.Hermes(hermes.backend.inprocess.Backend)

    @cache(tags = ('tag1', 'tag2'))
    def foo(a, b):
      return a * b

    foo(2, 2)

    print(cache.backend.dump())
    #  {
    #    'cache:tag:tag1': 'Z_t4-2QfjwvYUc_d75WJgw',
    #    'cache:tag:tag2': 'M_2pg1-dfS_8OM1Dnd4mXw',
    #    'cache:entry:foo:JjwE0zQhMRt_5lfwNNPk1Q:ZuT3KKj6gttXtU01-JSbwQ': 4
    #  }

    cache.clean(['tag1'])
    foo(2, 2)

    print(cache.backend.dump())
    #  {
    #    'cache:tag:tag1': 'TAuY_hhl0a22Nm7P_dWQRA',
    #    'cache:tag:tag2': 'M_2pg1-dfS_8OM1Dnd4mXw',
    #    'cache:entry:foo:JjwE0zQhMRt_5lfwNNPk1Q:St3ZHOzerXZxN0qkzXcLVw': 4,
    #    'cache:entry:foo:JjwE0zQhMRt_5lfwNNPk1Q:ZuT3KKj6gttXtU01-JSbwQ': 4
    #  }


class TestFacade(unittest.TestCase):

  def testBackendInstanceInstantiation(self):
    mangler = Mangler()
    backend = AbstractBackend(mangler)
    testee = Hermes(backend, mangler = mangler)

    @testee
    def mul(a, b):
      return a * b

    self.assertTrue(4, mul(2, 2))

  def testBackendInstanceWarning(self):
    mangler = Mangler()
    backend = AbstractBackend(mangler)

    with self.assertWarns(Warning) as ctx:
      testee = Hermes(backend, mangler = mangler, someBackendOpt = 13)
    self.assertEqual(
      'Backend options ignored because backend instance is passed', str(ctx.warning)
    )

    @testee
    def mul(a, b):
      return a * b

    self.assertTrue(4, mul(2, 2))

  def testInvalidInstantiation(self):
    with self.assertRaises(HermesError) as ctx:
      Hermes(object())
    self.assertEqual('Expected class or instance of AbstractBackend', str(ctx.exception))


class TestDefaultMangler(unittest.TestCase):

  testee: Mangler

  def setUp(self):
    self.testee = Mangler()

  def testHash(self):
    self.assertEqual('rL0Y20zC-Fzt72VPzMSk2A', self.testee.hash(b'foo'))

  def testSerialisation(self):
    value = {'meaning': '42' * 100}
    self.assertEqual(value, self.testee.loads(self.testee.dumps(value)))

  def testSerialisationCustom(self):
    self.testee.serialiser = Serialiser(lambda obj: json.dumps(obj).encode(), json.loads)

    value = {'meaning': '42' * 100}
    self.assertEqual(value, self.testee.loads(self.testee.dumps(value)))

  def testSerialisationCompressorOptOut(self):
    self.testee.compressor = None

    value = {'meaning': '42' * 100}
    self.assertEqual(value, self.testee.loads(self.testee.dumps(value)))

  def testSerialisationCompressorCustom(self):
    self.testee.compressor = Compressor(lzma.compress, lzma.decompress, lzma.LZMAError, 100)

    value = {'meaning': '42' * 100}
    self.assertEqual(value, self.testee.loads(self.testee.dumps(value)))

  def testNameEntryFunction(self):
    def fn(*args, **kwargs):
      pass

    no_arg_hash = self.testee.hash(self.testee.dumps(((), ())))
    pos_arg_hash = self.testee.hash(self.testee.dumps((('p',), ())))
    kw_arg_hash = self.testee.hash(self.testee.dumps(((), (('a', 1), ('k', 'v')))))
    both_arg_hash = self.testee.hash(self.testee.dumps((('p',), (('k', 'v'),))))

    prefix = 'cache:entry:hermes.test.facade'
    self.assertEqual(f'{prefix}:fn:{no_arg_hash}', self.testee.nameEntry(fn))
    self.assertEqual(f'{prefix}:fn:{pos_arg_hash}', self.testee.nameEntry(fn, 'p'))
    self.assertEqual(f'{prefix}:fn:{kw_arg_hash}', self.testee.nameEntry(fn, k = 'v', a = 1))
    self.assertEqual(f'{prefix}:fn:{both_arg_hash}', self.testee.nameEntry(fn, 'p', k = 'v'))

  def testNameEntryMethod(self):
    class Cls:

      def meth(self):
        pass

      @classmethod
      def clsmeth(cls):
        pass

      @staticmethod
      def statmeth():
        pass

    obj = Cls()
    prefix = 'cache:entry:hermes.test.facade'

    testfn = self.testee.nameEntry
    no_arg_hash = self.testee.hash(self.testee.dumps(((), ())))
    pos_arg_hash = self.testee.hash(self.testee.dumps((('p',), ())))
    kw_arg_hash = self.testee.hash(self.testee.dumps(((), (('a', 1), ('k', 'v')))))
    both_arg_hash = self.testee.hash(self.testee.dumps((('p',), (('k', 'v'),))))

    self.assertEqual(f'{prefix}:meth:{no_arg_hash}', testfn(Cls.meth))
    self.assertEqual(f'{prefix}:meth:{pos_arg_hash}', testfn(Cls.meth, 'p'))
    self.assertEqual(f'{prefix}:meth:{kw_arg_hash}', testfn(Cls.meth, k = 'v', a = 1))
    self.assertEqual(f'{prefix}:meth:{both_arg_hash}', testfn(Cls.meth, 'p', k = 'v'))

    self.assertEqual(f'{prefix}:type:clsmeth:{no_arg_hash}', testfn(Cls.clsmeth))
    self.assertEqual(f'{prefix}:type:clsmeth:{pos_arg_hash}', testfn(Cls.clsmeth, 'p'))
    self.assertEqual(f'{prefix}:type:clsmeth:{kw_arg_hash}', testfn(Cls.clsmeth, k = 'v', a = 1))
    self.assertEqual(f'{prefix}:type:clsmeth:{both_arg_hash}', testfn(Cls.clsmeth, 'p', k = 'v'))

    self.assertEqual(f'{prefix}:statmeth:{no_arg_hash}', testfn(Cls.statmeth))
    self.assertEqual(f'{prefix}:statmeth:{pos_arg_hash}', testfn(Cls.statmeth, 'p'))
    self.assertEqual(f'{prefix}:statmeth:{kw_arg_hash}', testfn(Cls.statmeth, k = 'v', a = 1))
    self.assertEqual(f'{prefix}:statmeth:{both_arg_hash}', testfn(Cls.statmeth, 'p', k = 'v'))

    self.assertEqual(f'{prefix}:Cls:meth:{no_arg_hash}', testfn(obj.meth))
    self.assertEqual(f'{prefix}:Cls:meth:{pos_arg_hash}', testfn(obj.meth, 'p'))
    self.assertEqual(f'{prefix}:Cls:meth:{kw_arg_hash}', testfn(obj.meth, k = 'v', a = 1))
    self.assertEqual(f'{prefix}:Cls:meth:{both_arg_hash}', testfn(obj.meth, 'p', k = 'v'))

    self.assertEqual(f'{prefix}:type:clsmeth:{no_arg_hash}', testfn(obj.clsmeth))
    self.assertEqual(f'{prefix}:type:clsmeth:{pos_arg_hash}', testfn(obj.clsmeth, 'p'))
    self.assertEqual(f'{prefix}:type:clsmeth:{kw_arg_hash}', testfn(obj.clsmeth, k = 'v', a = 1))
    self.assertEqual(f'{prefix}:type:clsmeth:{both_arg_hash}', testfn(obj.clsmeth, 'p', k = 'v'))

    self.assertEqual(f'{prefix}:statmeth:{no_arg_hash}', testfn(obj.statmeth))
    self.assertEqual(f'{prefix}:statmeth:{pos_arg_hash}', testfn(obj.statmeth, 'p'))
    self.assertEqual(f'{prefix}:statmeth:{kw_arg_hash}', testfn(obj.statmeth, k = 'v', a = 1))
    self.assertEqual(f'{prefix}:statmeth:{both_arg_hash}', testfn(obj.statmeth, 'p', k = 'v'))

  def testNameEntryCallableObject(self):
    class UnnamedCallable:

      def __call__(self, *args, **kwargs):
        pass

    prefix = 'cache:entry:hermes.test.facade'

    obj = UnnamedCallable()

    with self.assertRaises(HermesError) as ctx:
      self.testee.nameEntry(obj)
    self.assertEqual(
      'fn is callable but its name is undefined, consider overriding Mangler.nameEntry',
      str(ctx.exception),
    )

    class NamedCallable(UnnamedCallable):

      @property
      def __name__(self):
        return 'd.b'

    no_arg_hash = self.testee.hash(self.testee.dumps(((), ())))
    pos_arg_hash = self.testee.hash(self.testee.dumps((('p',), ())))
    kw_arg_hash = self.testee.hash(self.testee.dumps(((), (('a', 1), ('k', 'v')))))
    both_arg_hash = self.testee.hash(self.testee.dumps((('p',), (('k', 'v'),))))

    obj = NamedCallable()
    self.assertEqual(f'{prefix}:d.b:{no_arg_hash}', self.testee.nameEntry(obj))
    self.assertEqual(f'{prefix}:d.b:{pos_arg_hash}', self.testee.nameEntry(obj, 'p'))
    self.assertEqual(f'{prefix}:d.b:{kw_arg_hash}', self.testee.nameEntry(obj, k = 'v', a = 1))
    self.assertEqual(f'{prefix}:d.b:{both_arg_hash}', self.testee.nameEntry(obj, 'p', k = 'v'))

  def testNameEntryUncallable(self):
    with self.assertRaises(HermesError) as ctx:
      self.testee.nameEntry({'foo': 'bar'})
    self.assertEqual('fn is expected to be callable', str(ctx.exception))

  def testNameTag(self):
    self.assertEqual('cache:tag:foo', self.testee.nameTag('foo'))

  def testMapTags(self):
    self.assertEqual({}, self.testee.mapTags([]))

    actual = self.testee.mapTags(['foo', 'bar'])
    self.assertEqual(2, len(actual))
    self.assertIn('foo', actual)
    self.assertIn('bar', actual)
    self.assertNotEqual(actual, self.testee.mapTags({'foo'}))

  def testHashTags(self):
    # Normally a combination of nameTag() and mapTags()
    tagMap = {'cache:tag:a': '2209aaab3ace901e', 'cache:tag:z': '92670722746eda09'}
    self.assertEqual('Iq0ue2khRbyyTyLZJecJEA', self.testee.hashTags(tagMap))

  def testNameLock(self):
    self.assertEqual('cache:lock:foo.bar', self.testee.nameLock('foo.bar'))
    self.assertEqual(
      'cache:lock:hermes.test:Fixture:tagged:78d64ea049a57494',
      self.testee.nameLock('cache:entry:hermes.test:Fixture:tagged:78d64ea049a57494'),
    )


class TestCachedWrapping(test.TestCase):

  def setUp(self):
    mangler = Mangler()
    backend = inprocess.Backend(mangler, ttlWatchSleep = 0.001)
    self.testee = Hermes(backend, mangler = mangler)
    self.addCleanup(self.testee.backend.stopWatch)
    self.fixture = test.createFixture(self.testee)

  def testFunction(self):

    @self.testee
    def foo(a, b):
      '''Overwhelmed as one would be...'''

      return a * b

    self.assertTrue(isinstance(foo, Cached))
    self.assertEqual('foo', foo.__name__)
    self.assertEqual('Overwhelmed as one would be...', foo.__doc__)

  def testMethod(self):
    self.assertTrue(isinstance(self.fixture.simple, Cached))
    self.assertEqual('simple', self.fixture.simple.__name__)
    self.assertEqual(
      'Here be dragons... seriously just a docstring test.', self.fixture.simple.__doc__)

  def testInstanceIsolation(self):
    testee = Hermes()

    class Fixture:

      def __init__(self, marker):
        self.marker = marker

      @testee
      def foo(self):
        return self.marker

      def bar(self):
        pass

    f1 = Fixture(12)
    f2 = Fixture(24)

    # verify Cached instances are not shared
    self.assertIsNot(f1.foo, f2.foo)
    self.assertIsNot(f1.foo, f1.foo)
    # like it is normally the case of MethodType instances
    self.assertIsNot(f1.bar, f2.bar)
    self.assertIsNot(f1.bar, f1.bar)

    self.assertEqual(12, f1.foo())
    self.assertEqual(24, f2.foo())

  def testMethodDescriptor(self):

    class Fixture:

      @self.testee
      @classmethod
      def classmethod(cls):
        return cls.__name__

      @self.testee
      @staticmethod
      def staticmethod():
        return 'static'

    self.assertEqual('Fixture', Fixture().classmethod())
    self.assertEqual('static', Fixture().staticmethod())

  @asynctest
  async def testCoroutineFunction(self):
    testee = Hermes(inprocess.AsyncBackend)

    @testee
    async def foo(a, b):
      '''Placed in my position...'''

      return a * b

    self.assertTrue(isinstance(foo, CachedCoro))
    self.assertEqual('foo', foo.__name__)
    self.assertEqual('Placed in my position...', foo.__doc__)
    self.assertEqual(4, await foo(2, 2))

  @asynctest
  async def testMethodDescriptorCoroutineFunction(self):
    testee = Hermes(inprocess.AsyncBackend)

    class Fixture:

      @testee
      @classmethod
      def sclassmethod(cls):
        return cls.__name__

      @testee
      @staticmethod
      def sstaticmethod():
        return 'static'

      @testee
      @classmethod
      async def aclassmethod(cls):
        return cls.__name__[::-1]

      @testee
      @staticmethod
      async def astaticmethod():
        return 'citats'

    fixture = Fixture()
    self.assertEqual('Fixture', fixture.sclassmethod())
    self.assertEqual('static', fixture.sstaticmethod())
    self.assertEqual('erutxiF', await fixture.aclassmethod())
    self.assertEqual('citats', await fixture.astaticmethod())

    self.assertIsInstance(fixture.sclassmethod, Cached)
    self.assertIsInstance(fixture.sstaticmethod, Cached)
    self.assertIsInstance(fixture.aclassmethod, CachedCoro)
    self.assertIsInstance(fixture.astaticmethod, CachedCoro)

  def testDoubleCache(self):

    class Fixture:

      @self.testee
      @self.testee
      def doublecache(self):
        return 'descriptor vs callable'

    self.assertEqual('descriptor vs callable', Fixture().doublecache())

  def testNumbaCpuDispatcher(self):
    # In case of application of ``jit`` to a method, ``CpuDispatcher`` is
    # a descriptor which returns normal MethodType. In case of function,
    # ``CpuDispatcher`` is a callable and has ``__name__`` of the wrapped function.

    class CpuDispatcher:

      def __init__(self, fn):
        self.fn = fn

      @property
      def __name__(self):
        return self.fn.__name__

      def __call__(self, *args, **kwargs):
        return self.fn(*args, **kwargs)

    def jit(fn):
      return CpuDispatcher(fn)

    @self.testee
    @jit
    def sum(x, y):
      return x + y

    @self.testee
    @jit
    def product(x, y):
      return x * y

    self.assertEqual(36, sum(26, 10))
    self.assertEqual(260, product(26, 10))

  def testNamelessObjectWrapperFailure(self):

    class CpuDispatcher:

      def __init__(self, fn):
        self.fn = fn

      def __call__(self, *args, **kwargs):
        return self.fn(*args, **kwargs)

    def objwrap(fn):
      return CpuDispatcher(fn)

    @self.testee
    @objwrap
    def f(x, y):
      return x + y

    with self.assertRaises(HermesError) as ctx:
      f(26, 10)
    self.assertEqual(
      'fn is callable but its name is undefined, consider overriding Mangler.nameEntry',
      str(ctx.exception))

  def testUncallable(self):
    with self.assertRaises(HermesError) as ctx:
      self.testee({'meaning': 42})
    self.assertEqual(
      'First positional argument must be coroutine function, callable or method descriptor',
      str(ctx.exception),
    )

  def testClassAccess(self):
    self.assertIsInstance(self.fixture.__class__.simple, Cached)
    self.assertIsInstance(self.fixture.__class__.simple._callable, types.FunctionType)

    self.assertIsInstance(self.fixture.simple, Cached)
    self.assertIsInstance(self.fixture.simple._callable, types.MethodType)

  def testInvalidate(self):
    for _ in range(4):
      self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))

    self.fixture.simple.invalidate('beta', 'gamma')
    self.assertNotEqual({}, self.testee.backend.dump())

    self.fixture.simple.invalidate('alpha', 'beta')
    self.assertEqual({}, self.testee.backend.dump())

  def testInvalidateFailPartialTags(self):
    for _ in range(4):
      self.assertEqual('ae-hl', self.fixture.tagged('alpha', 'beta'))

    # Simulate missing tag entry
    self.testee.backend.remove('cache:tag:tree')

    # Invalidation call does not have effect
    self.fixture.tagged.invalidate('alpha', 'beta')

    self.assertNotEqual({}, self.testee.backend.dump())
    self.assertEqual(2, len(self.testee.backend.dump()))

    # Recover missing tag
    for _ in range(4):
      self.assertEqual('ae-hl', self.fixture.tagged('alpha', 'beta'))

    self.fixture.tagged.invalidate('alpha', 'beta')
    self.assertNotEqual({}, self.testee.backend.dump())


class CustomMangler(Mangler):

  prefix = 'hermes'

  compressor = Compressor(zlib.compress, zlib.decompress, zlib.error, 0)

  def hash(self, value):
    return str(hash(value))


class TestDictCustomMangler(test.TestCase):

  def setUp(self):
    self.testee = Hermes(
      inprocess.Backend, mangler = CustomMangler(), ttl = 360, ttlWatchSleep = 0.01
    )
    self.addCleanup(self.testee.backend.stopWatch)

    self.fixture = test.createFixture(self.testee)

    self.testee.clean()

  def testSimple(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

    for _ in range(4):
      self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
      self.assertEqual(1, self.fixture.calls)
      key = 'hermes:entry:hermes.test:Fixture:simple:' + str(self._arghash('alpha', 'beta'))
      self.assertEqual({key: 'ateb+ahpla'}, self.testee.backend.dump())

    self.fixture.simple.invalidate('alpha', 'beta')
    self.assertEqual({}, self.testee.backend.dump())


    self.assertEqual(1,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

    expected = "]}'atled' :'ammag'{[+}]'ateb'[ :'ahpla'{"
    for _ in range(4):
      self.assertEqual(expected, self.fixture.simple({'alpha': ['beta']}, [{'gamma': 'delta'}]))
      self.assertEqual(2, self.fixture.calls)
      argHash = str(self._arghash({'alpha': ['beta']}, [{'gamma': 'delta'}]))
      self.assertEqual(
        {'hermes:entry:hermes.test:Fixture:simple:' + argHash: expected},
        self.testee.backend.dump())

    self.fixture.simple.invalidate({'alpha': ['beta']}, [{'gamma': 'delta'}])
    self.assertEqual({}, self.testee.backend.dump())

  def testTagged(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

    for _ in range(4):
      self.assertEqual('ae-hl', self.fixture.tagged('alpha', 'beta'))
      self.assertEqual(1,       self.fixture.calls)

      cache = self.testee.backend.dump()
      self.assertTrue(3, len(cache))
      self.assertFalse(cache.get('hermes:tag:tree') == cache.get('hermes:tag:rock'))
      self.assertTrue(int(cache.pop('hermes:tag:rock')) != 0)
      self.assertTrue(int(cache.pop('hermes:tag:tree')) != 0)

      expected = 'hermes:entry:hermes.test:Fixture:tagged:' + str(self._arghash('alpha', 'beta'))
      self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
      self.assertEqual('ae-hl', tuple(cache.values())[0])

    self.fixture.tagged.invalidate('alpha', 'beta')

    cache = self.testee.backend.dump()
    self.assertEqual(2, len(cache))
    rockTag = cache.get('hermes:tag:rock')
    treeTag = cache.get('hermes:tag:tree')
    self.assertNotEqual(rockTag, treeTag)
    self.assertTrue(int(rockTag) != 0)
    self.assertTrue(int(treeTag) != 0)

    for _ in range(4):
      self.assertEqual('ae-hl', self.fixture.tagged('alpha', 'beta'))
      self.assertEqual('ae%hl', self.fixture.tagged2('alpha', 'beta'))
      self.assertEqual(3, self.fixture.calls)

      cache = self.testee.backend.dump()
      self.assertEqual(5,  len(cache))
      self.assertEqual(rockTag, cache.get('hermes:tag:rock'))
      self.assertEqual(treeTag, cache.get('hermes:tag:tree'))
      self.assertTrue(int(cache.get('hermes:tag:ice')) != 0)

    self.testee.clean(['rock'])

    cache = self.testee.backend.dump()
    self.assertEqual(4, len(cache))
    self.assertTrue('hermes:tag:rock' not in cache)
    iceTag = cache.get('hermes:tag:ice')

    for _ in range(4):
      self.assertEqual('ae-hl', self.fixture.tagged('alpha', 'beta'))
      self.assertEqual('ae%hl', self.fixture.tagged2('alpha', 'beta'))
      self.assertEqual(5, self.fixture.calls)

      cache = self.testee.backend.dump()
      self.assertEqual(7,  len(cache), 'has new and old entries for tagged and tagged 2 + 3 tags')
      self.assertEqual(treeTag, cache.get('hermes:tag:tree'))
      self.assertEqual(iceTag,  cache.get('hermes:tag:ice'))
      self.assertTrue(int(cache.get('hermes:tag:rock')) != 0)
      self.assertNotEqual(rockTag, cache.get('hermes:tag:rock'))

  def testFunction(self):
    counter = dict(foo = 0, bar = 0)

    @self.testee
    def foo(a, b):
      counter['foo'] += 1
      return '{0}+{1}'.format(a, b)[::-1]

    key = lambda _fn, *args, **kwargs: 'mk:{0}:{1}'.format(*args)

    @self.testee(tags = ('a', 'z'), key = key, ttl = 120)
    def bar(a, b):
      counter['bar'] += 1
      return '{0}-{1}'.format(a, b)[::2]


    self.assertEqual(0,  counter['foo'])
    self.assertEqual({}, self.testee.backend.dump())

    for _ in range(4):
      self.assertEqual('ateb+ahpla', foo('alpha', 'beta'))
      self.assertEqual(1, counter['foo'])
      argHash = self._arghash('alpha', 'beta')
      self.assertEqual(
        {'hermes:entry:hermes.test.facade:foo:' + str(argHash): 'ateb+ahpla'},
        self.testee.backend.dump()
      )

    foo.invalidate('alpha', 'beta')
    self.assertEqual(1,  counter['foo'])
    self.assertEqual({}, self.testee.backend.dump())


    self.assertEqual(0,  counter['bar'])
    self.assertEqual({}, self.testee.backend.dump())

    for _ in range(4):
      self.assertEqual('apabt', bar('alpha', 'beta'))
      self.assertEqual(1,       counter['bar'])

      cache = self.testee.backend.dump()
      self.assertTrue(cache.get('hermes:tag:a') != cache.get('hermes:tag:z'))
      self.assertTrue(int(cache.pop('hermes:tag:a')) != 0)
      self.assertTrue(int(cache.pop('hermes:tag:z')) != 0)

      self.assertEqual('mk:alpha:beta', ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
      self.assertEqual('apabt', tuple(cache.values())[0])

    bar.invalidate('alpha', 'beta')


    self.assertEqual(1,  counter['foo'])

    cache = self.testee.backend.dump()
    self.assertTrue(cache.get('hermes:tag:a') != cache.get('hermes:tag:z'))
    self.assertTrue(int(cache.pop('hermes:tag:a')) != 0)
    self.assertTrue(int(cache.pop('hermes:tag:z')) != 0)

  def testKey(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

    for _ in range(4):
      self.assertEqual('apabt', self.fixture.key('alpha', 'beta'))
      self.assertEqual(1,       self.fixture.calls)

      cache = self.testee.backend.dump()
      self.assertTrue(cache.get('hermes:tag:ash') != cache.get('hermes:tag:stone'))
      self.assertTrue(int(cache.pop('hermes:tag:ash')) != 0)
      self.assertTrue(int(cache.pop('hermes:tag:stone')) != 0)

      self.assertEqual('mykey:alpha:beta', ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
      self.assertEqual('apabt', tuple(cache.values())[0])

    self.fixture.key.invalidate('alpha', 'beta')

    cache = self.testee.backend.dump()
    self.assertTrue(cache.get('hermes:tag:ash') != cache.get('hermes:tag:stone'))
    self.assertTrue(int(cache.pop('hermes:tag:ash')) != 0)
    self.assertTrue(int(cache.pop('hermes:tag:stone')) != 0)

  def testAll(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

    for _ in range(4):
      self.assertEqual({'a': 1, 'b': {'b': 'beta'}}, self.fixture.all({'alpha': 1}, ['beta']))
      self.assertEqual(1, self.fixture.calls)

      cache = self.testee.backend.dump()
      self.assertEqual(3, len(cache))
      self.assertTrue(int(cache.pop('hermes:tag:a')) != 0)
      self.assertTrue(int(cache.pop('hermes:tag:z')) != 0)

      expected = "mk:{'alpha':1}:['beta']"
      self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
      self.assertEqual({'a': 1, 'b': {'b': 'beta'}}, tuple(cache.values())[0])

    self.fixture.all.invalidate({'alpha': 1}, ['beta'])

    cache = self.testee.backend.dump()
    self.assertEqual(2, len(cache))
    self.assertTrue(int(cache.pop('hermes:tag:a')) != 0)
    self.assertTrue(int(cache.pop('hermes:tag:z')) != 0)

  def testClean(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(2,            self.fixture.calls)

    self.testee.clean()

    self.assertEqual(2,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

  def testCleanTagged(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(2,            self.fixture.calls)

    cache = self.testee.backend.dump()
    self.assertTrue(4, len(cache))
    self.assertEqual(
      'ateb+ahpla',
      cache.pop(
        'hermes:entry:hermes.test:Fixture:simple:' + str(self._arghash('alpha', 'beta'))
      ),
    )
    self.assertFalse(cache.get('hermes:tag:tree') == cache.get('hermes:tag:rock'))
    self.assertTrue(int(cache.pop('hermes:tag:rock')) != 0)
    self.assertTrue(int(cache.pop('hermes:tag:tree')) != 0)

    expected = 'hermes:entry:hermes.test:Fixture:tagged:' + str(self._arghash('gamma', 'delta'))
    self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('aldamg', tuple(cache.values())[0])


    self.testee.clean(('rock',))

    cache = self.testee.backend.dump()
    self.assertTrue(3, len(cache))
    self.assertEqual(
      'ateb+ahpla',
      cache.pop(
        'hermes:entry:hermes.test:Fixture:simple:' + str(self._arghash('alpha', 'beta'))
      ),
    )
    self.assertTrue(int(cache.pop('hermes:tag:tree')) != 0)
    self.assertFalse('hermes:tag:rock' in cache)

    self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('aldamg', tuple(cache.values())[0])


    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(3,            self.fixture.calls)

    cache = self.testee.backend.dump()
    self.assertTrue(4, len(cache))
    self.assertEqual(
      'ateb+ahpla',
      cache.pop(
        'hermes:entry:hermes.test:Fixture:simple:' + str(self._arghash('alpha', 'beta'))
      ),
    )
    self.assertFalse(cache.get('hermes:tag:tree') == cache.get('hermes:tag:rock'))
    self.assertTrue(int(cache.pop('hermes:tag:rock')) != 0)
    self.assertTrue(int(cache.pop('hermes:tag:tree')) != 0)

    self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('aldamg', tuple(cache.values())[0])


    self.testee.clean(('rock', 'tree'))

    cache = self.testee.backend.dump()
    self.assertTrue(2, len(cache))
    self.assertEqual(
      'ateb+ahpla',
      cache.pop(
        'hermes:entry:hermes.test:Fixture:simple:' + str(self._arghash('alpha', 'beta'))
      ),
    )
    self.assertFalse('hermes:tag:tree' in cache)
    self.assertFalse('hermes:tag:rock' in cache)

    self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('aldamg', tuple(cache.values())[0])


    self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
    self.assertEqual('aldamg',     self.fixture.tagged('gamma', 'delta'))
    self.assertEqual(4,            self.fixture.calls)

    cache = self.testee.backend.dump()
    self.assertTrue(4, len(cache))
    self.assertEqual(
      'ateb+ahpla',
      cache.pop(
        'hermes:entry:hermes.test:Fixture:simple:' + str(self._arghash('alpha', 'beta'))
      ),
    )
    self.assertFalse(cache.get('hermes:tag:tree') == cache.get('hermes:tag:rock'))
    self.assertTrue(int(cache.pop('hermes:tag:rock')) != 0)
    self.assertTrue(int(cache.pop('hermes:tag:tree')) != 0)

    self.assertEqual(expected, ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('aldamg', tuple(cache.values())[0])


    self.testee.clean()

    self.assertEqual(4,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

  def testNested(self):
    self.assertEqual('beta+alpha', self.fixture.nested('alpha', 'beta'))
    self.assertEqual(2, self.fixture.calls)
    prefix = 'hermes:entry:hermes.test:Fixture'
    self.assertEqual({
      prefix + ':nested:' + str(self._arghash('alpha', 'beta')): 'beta+alpha',
      prefix + ':simple:' + str(self._arghash('beta', 'alpha')): 'ahpla+ateb'
    }, self.testee.backend.dump())

  def testConcurrent(self):
    log = []
    key = lambda _fn, *args, **kwargs: 'mk:{0}:{1}'.format(*args)

    @self.testee(tags = ('a', 'z'), key = key, ttl = 120)
    def bar(a, b):
      log.append(1)
      time.sleep(0.04)
      return '{0}-{1}'.format(a, b)[::2]

    threads = [threading.Thread(target = bar, args = ('alpha', 'beta')) for _ in range(4)]
    tuple(map(threading.Thread.start, threads))
    tuple(map(threading.Thread.join,  threads))

    self.assertEqual(1, sum(log))

    cache = self.testee.backend.dump()
    self.assertTrue(int(cache.pop('hermes:tag:a')) != 0)
    self.assertTrue(int(cache.pop('hermes:tag:z')) != 0)

    self.assertEqual('mk:alpha:beta', ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('apabt', tuple(cache.values())[0])

    del log[:]
    self.testee.clean()
    self.testee.backend.lock = lambda k: AbstractLock(k)  # now see a dogpile

    threads = [threading.Thread(target = bar, args = ('alpha', 'beta')) for _ in range(4)]
    tuple(map(threading.Thread.start, threads))
    tuple(map(threading.Thread.join,  threads))

    self.assertGreater(sum(log), 1, 'dogpile')

    cache = self.testee.backend.dump()
    self.assertTrue(int(cache.pop('hermes:tag:a')) != 0)
    self.assertTrue(int(cache.pop('hermes:tag:z')) != 0)

    self.assertEqual('mk:alpha:beta', ':'.join(tuple(cache.keys())[0].split(':')[:-1]))
    self.assertEqual('apabt', tuple(cache.values())[0])


class CustomCached(Cached):

  def __call__(self, *args, **kwargs):
    ''''Override to add bypass when backend is down.'''

    try:
      return super().__call__(*args, **kwargs)
    except RuntimeError:
      return self._callable(*args, **kwargs)


class BackendThatIsDown(inprocess.Backend):

  def load(self, keys):
    raise RuntimeError('Backend is down')


class TestDictCustomCached(test.TestCase):

  def setUp(self):
    self.testee = Hermes(
      BackendThatIsDown, cachedfactory = self.cachedfactory, ttl = 360, ttlWatchSleep = 0.01
    )
    self.fixture = test.createFixture(self.testee)

    self.testee.clean()

  def cachedfactory(self, frontend: Hermes, fn, **kwargs) -> Cached:
    if inspect.iscoroutinefunction(fn):
      return CachedCoro(frontend, fn, **kwargs)
    elif callable(fn) or inspect.ismethoddescriptor(fn):
      return CustomCached(frontend, fn, **kwargs)
    else:
      raise HermesError(
        'First positional argument must be coroutine function, callable or method descriptor'
      )

  def testSimple(self):
    self.assertEqual(0,  self.fixture.calls)
    self.assertEqual({}, self.testee.backend.dump())

    for i in range(4):
      self.assertEqual('ateb+ahpla', self.fixture.simple('alpha', 'beta'))
      self.assertEqual(i + 1, self.fixture.calls)
      self.assertEqual({}, self.testee.backend.dump())



import asyncio
import base64
import functools
import hashlib
import inspect
import os
import pickle
import types
import typing
import warnings
import zlib
from typing import Any, Callable, Coroutine, Dict, Iterable, Optional, Sequence, Tuple, Type, Union

from .backend import AbstractBackend


__all__ = 'Hermes', 'HermesError', 'Mangler', 'Cached', 'CachedCoro', 'Serialiser', 'Compressor'


class Serialiser(typing.NamedTuple):
  '''Serialisation delegate.'''

  dumps: Callable[[Any], bytes]
  '''Serialise cache value.'''

  loads: Callable[[bytes], Any]
  '''Deserialise cache value.'''


class Compressor(typing.NamedTuple):
  '''Compression delegate.'''

  compress: Callable[[bytes], bytes]
  '''Compress serialised cache value.'''

  decompress: Callable[[bytes], bytes]
  '''Decompress serialised cache value.'''

  decompressError: Union[Type[Exception], Tuple[Type[Exception], ...]]
  '''Decompression error(s) that indicate uncompressed payload.'''

  compressMinLength: int = 0
  '''Minimal length of payload in bytes to trigger compression.'''


class Mangler:
  '''Key manager responsible for creating keys, hashing and serialisation.'''

  prefix = 'cache'
  '''Prefix for cache and tag entries.'''

  serialiser = Serialiser(pickle.dumps, pickle.loads)
  '''Serialisation delegate.'''

  compressor = Compressor(zlib.compress, zlib.decompress, zlib.error, 100)
  '''Optional compression delegate.'''

  def hash(self, value: bytes) -> str:
    '''
    Hash value.

    :return: base64 encoded MD5 hash of the value.
    '''

    return base64.urlsafe_b64encode(hashlib.md5(value).digest()).strip(b'=').decode()

  def dumps(self, value) -> bytes:
    '''Serialise and conditionally compress value.'''

    result = self.serialiser.dumps(value)
    if self.compressor and len(result) >= self.compressor.compressMinLength:
      result = self.compressor.compress(result)

    return result

  def loads(self, value: bytes):
    '''Conditionally decompress and deserialise value.'''

    if self.compressor:
      try:
        value = self.compressor.decompress(value)
      except self.compressor.decompressError:
        # It is expected that the error indicates that the value is
        # shorter than compressMinLength
        pass

    return self.serialiser.loads(value)

  def nameEntry(self, fn: Callable, *args, **kwargs) -> str:
    '''
    Return cache key for given callable and its positional and
    keyword arguments.

    Note how callable, ``fn``, is represented in the cache key:

      1) a ``types.MethodType`` instance -> names of
         ``(module, class, method)``
      2) a ``types.FunctionType`` instance -> names of
         ``(module, function)``
      3) other callalbe objects with ``__name__`` -> name of
         ``(module, object)``

    This means that if two function are defined dynamically in the
    same module with same names, like::

      def createF1():
          @cache
          def f(a, b):
              return a + b
          return f

      def createF2():
          @cache
          def f(a, b):
              return a * b
          return f

      print(createF1()(1, 2))
      print(createF2()(1, 2))

    Both will return `3`, because cache keys will clash. In such cases
    you need to pass ``key`` with custom key function.

    It can also be that an object in case 3 doesn't have name, or its
    name isn't unique, then a ``nameEntry`` should be overridden with
    something that represents it uniquely, like
    ``repr(fn).rsplit(' at 0x', 1)[0]`` (address should be stripped so
    after Python process restart the cache can still be valid
    and usable).
    '''

    result = [self.prefix, 'entry']
    if callable(fn):
      try:
        # types.MethodType
        result.extend([
          fn.__module__,
          fn.__self__.__class__.__name__,  # type: ignore[attribute-error]
          fn.__name__,
        ])
      except AttributeError:
        try:
          # types.FunctionType and other object with __name__
          result.extend([fn.__module__, fn.__name__])
        except AttributeError:
          raise HermesError(
            'fn is callable but its name is undefined, consider overriding Mangler.nameEntry'
          )
    else:
      raise HermesError('fn is expected to be callable')

    arguments = args, tuple(sorted(kwargs.items()))
    result.append(self.hash(self.dumps(arguments)))

    return ':'.join(result)

  def nameTag(self, tag: str) -> str:
    '''Build fully qualified backend tag name.'''

    return ':'.join([self.prefix, 'tag', tag])

  def mapTags(self, tagKeys: Iterable[str]) -> Dict[str, str]:
    '''Map tags to random values for seeding.'''

    rnd = os.urandom(4).hex()
    return {key: self.hash(':'.join((key, rnd)).encode()) for key in tagKeys}

  def hashTags(self, tagMap: Dict[str, str]) -> str:
    '''Hash tags of a cache entry for the entry key,'''

    values = tuple(zip(*sorted(tagMap.items())))[1]  # sorted by key dict values
    return self.hash(':'.join(values).encode())

  def nameLock(self, entryKey: str) -> str:
    '''
    Create fully qualified backend lock key for the entry key.

    :param entryKey:
      Entry key to create a lock key for. If given entry key is already
      a colon-separated key name with first component equal to
      :attr:`prefix`, first to components are dropped. For instance:

      - ``foo`` → ``cache:lock:foo``
      - ``cache:entry:fn:tagged:78d64ea049a57494`` →
        ``cache:lock:fn:tagged:78d64ea049a57494``

    '''

    parts = entryKey.split(':')
    if parts[0] == self.prefix:
      entryKey = ':'.join(parts[2:])

    return ':'.join([self.prefix, 'lock', entryKey])


class Cached:
  '''Cache-point wrapper for callables and descriptors.'''

  _frontend: 'Hermes'
  '''
  Hermes instance which provides backend and mangler instances, and
  TTL fallback value.
  '''

  _callable: Callable
  '''
  The decorated callable, stays ``types.FunctionType`` if a function
  is decorated, otherwise it is transformed to ``types.MethodType``
  on the instance clone by descriptor protocol implementation. It can
  also be a method descriptor which is also transformed accordingly to
  the descriptor protocol (e.g. ``staticmethod`` and ``classmethod``).
  '''

  _isDescriptor: bool
  '''Flag defining if the callable is a method descriptor.'''

  _isMethod: bool
  '''Flag defining if the callable is a method.'''

  _ttl: Optional[int]
  '''Cache entry Time To Live for decorated callable.'''

  _keyFunc: Optional[Callable]
  '''Key creation function.'''

  _tags: Sequence[str]
  '''Cache entry tags for decorated callable.'''

  def __init__(
    self,
    frontend: 'Hermes',
    callable: Callable,
    *,
    ttl: Optional[int] = None,
    key: Optional[Callable] = None,
    tags: Sequence[str] = (),
  ):
    self._frontend = frontend
    self._ttl      = ttl
    self._keyFunc  = key
    self._tags     = tags

    self._callable     = callable
    self._isDescriptor = inspect.ismethoddescriptor(callable)
    self._isMethod     = inspect.ismethod(callable)

    # preserve ``__name__``, ``__doc__``, etc
    functools.update_wrapper(self, callable)

  def _load(self, key):
    if self._tags:
      tagMap = self._frontend.backend.load(map(self._frontend.mangler.nameTag, self._tags))
      if len(tagMap) != len(self._tags):
        return None
      else:
        key += ':' + self._frontend.mangler.hashTags(tagMap)

    return self._frontend.backend.load(key)

  def _save(self, key, value):
    if self._tags:
      namedTags   = tuple(map(self._frontend.mangler.nameTag, self._tags))
      tagMap      = self._frontend.backend.load(namedTags)
      missingTags = set(namedTags) - set(tagMap.keys())
      if missingTags:
        missingTagMap = self._frontend.mangler.mapTags(missingTags)
        self._frontend.backend.save(mapping = missingTagMap, ttl = None)
        tagMap.update(missingTagMap)
        assert len(self._tags) == len(tagMap)

      key += ':' + self._frontend.mangler.hashTags(tagMap)

    ttl = self._ttl if self._ttl is not None else self._frontend.ttl

    return self._frontend.backend.save({key: value}, ttl = ttl)

  def _remove(self, key):
    if self._tags:
      tagMap = self._frontend.backend.load(map(self._frontend.mangler.nameTag, self._tags))
      if len(tagMap) != len(self._tags):
        return
      else:
        key += ':' + self._frontend.mangler.hashTags(tagMap)

    self._frontend.backend.remove(key)

  def _key(self, *args, **kwargs):
    keyFunc = self._keyFunc or self._frontend.mangler.nameEntry
    return keyFunc(self._callable, *args, **kwargs)

  def invalidate(self, *args, **kwargs):
    '''
    Invalidate the cache entry.

    Invalidated entry corresponds to the wrapped callable called with
    given ``args`` and ``kwargs``.
    '''

    self._remove(self._key(*args, **kwargs))

  def __call__(self, *args, **kwargs):
    '''Get the value of the wrapped callable.'''

    key   = self._key(*args, **kwargs)
    value = self._load(key)
    if value is None:
      with self._frontend.backend.lock(key):
        # it's better to read twice than lock every read
        value = self._load(key)
        if value is None:
          value = self._callable(*args, **kwargs)
          self._save(key, value)

    return value

  def __get__(self, instance, type):
    '''
    Implements non-data descriptor protocol.

    The invocation happens only when instance method is decorated,
    so we can distinguish between decorated ``types.MethodType`` and
    ``types.FunctionType``. Python class declaration mechanics prevent
    a decorator from having awareness of the class type, as the
    function is received by the decorator before it becomes an
    instance method.

    How it works::

      cache = hermes.Hermes()

      class Model:

        @cache
        def calc(self):
          return 42

      m = Model()
      m.calc

    Last attribute access results in the call, ``calc.__get__(m, Model)``,
    where ``calc`` is instance of :class:`Cached` which decorates the
    original ``Model.calc``.

    Note, initially :class:`Cached` is created on decoration per
    class method, when class type is created by the interpreter, and
    is shared among all instances. Later, on attribute access, a copy
    is returned with bound ``_callable``, just like ordinary Python
    method descriptor works.

    For more details, `descriptor-protocol
    <http://docs.python.org/3/howto/descriptor.html#descriptor-protocol>`_.
    '''

    if instance is not None and self._isDescriptor:
      return self._copy(self._callable.__get__(instance, type))  # type: ignore[attribute-error]
    elif instance is not None and not self._isMethod:
      return self._copy(types.MethodType(self._callable, instance))
    else:
      return self

  def _copy(self, callable):
    '''
    Create a shallow copy of self with ``_callable``
    replaced to given instance.
    '''

    boundCached           = object.__new__(self.__class__)
    boundCached.__dict__  = self.__dict__.copy()
    boundCached._callable = callable
    return boundCached


class CachedCoro(Cached):
  '''
  Cache-point wrapper for coroutine functions.

  The implementation uses the default thread pool of ``asyncio`` to
  execute synchronous functions of the cache backend, and manage their
  (distributed) locks.
  '''

  async def _run(self, fn, *args, **kwargs) -> Coroutine:
    ''''
    Run run given function or coroutine function.

    If ``fn`` is a coroutine function it's called and awaited.
    Otherwise it's run in the thread pool.
    '''

    if inspect.iscoroutinefunction(fn):
      return await fn(*args, **kwargs)

    loop = asyncio.get_event_loop()
    return await loop.run_in_executor(None, functools.partial(fn, *args, **kwargs))

  async def invalidate(self, *args, **kwargs):
    '''
    Invalidate the cache entry.

    Invalidated entry corresponds to the wrapped coroutine function
    called with given ``args`` and ``kwargs``.
    '''

    await self._run(super().invalidate, *args, **kwargs)

  async def __call__(self, *args, **kwargs):
    '''Get the value of the wrapped coroutine function's coroutine.'''

    key = self._key(*args, **kwargs)
    value = await self._run(self._load, key)
    if value is None:
      lock = self._frontend.backend.lock(key)
      await self._run(lock.acquire)
      try:
        value = await self._run(self._load, key)
        if value is None:
          value = await self._callable(*args, **kwargs)
          await self._run(self._save, key, value)
      finally:
        await self._run(lock.release)

    return value


def cachedfactory(frontend: 'Hermes', fn, **kwargs) -> Cached:
  '''
  Create a cache-point object from the callable.

  :argument frontend:
    Cache frontend instance.
  :argument fn:
    Must be coroutine function, callable or method descriptor.
  '''

  isdescr = inspect.ismethoddescriptor(fn)
  if (
    inspect.iscoroutinefunction(fn)
    or isdescr and inspect.iscoroutinefunction(getattr(fn, '__func__', None))
  ):
    return CachedCoro(frontend, fn, **kwargs)
  elif callable(fn) or isdescr:
    return Cached(frontend, fn, **kwargs)
  else:
    raise HermesError(
      'First positional argument must be coroutine function, callable or method descriptor'
    )


class Hermes:
  '''
  Cache façade.

  :argument backend:
    Class or instance of cache backend. If a class is passed, keyword
    arguments of passed to :obj:`Hermes` constructor will be bypassed
    to the class' constructor.

    If the argument is omitted no-op backend will be be used.
  :argument mangler:
    Optional, typically of a subclass, mangler instance.
  :argument cachedfactory:
    Optional, a cache-point factory for functions and coroutines.
  :argument ttl:
    Default cache entry time-to-live.

  Usage::

    import hermes.backend.redis


    cache = hermes.Hermes(
      hermes.backend.redis.Backend, ttl = 600, host = 'localhost', db = 1
    )

    @cache
    def foo(a, b):
      return a * b

    class Example:

      @cache(tags = ('math', 'power'), ttl = 1200)
      def bar(self, a, b):
        return a ** b

      @cache(
        tags = ('math', 'avg'),
        key = lambda fn, *args, **kwargs: 'avg:{0}:{1}'.format(*args),
      )
      def baz(self, a, b):
        return (a + b) / 2.0

    print(foo(2, 333))

    example = Example()
    print(example.bar(2, 10))
    print(example.baz(2, 10))

    foo.invalidate(2, 333)
    example.bar.invalidate(2, 10)
    example.baz.invalidate(2, 10)

    cache.clean(['math']) # invalidate entries tagged 'math'
    cache.clean()         # flush cache

  '''

  backend: AbstractBackend
  '''Cache backend.'''

  mangler: Mangler
  '''Key manager responsible for creating keys, hashing and serialisation.'''

  cachedfactory: Callable[..., Cached]
  '''Cache-point callable object factory.'''

  ttl: int
  '''Default cache entry time-to-live.'''

  def __init__(
    self,
    backend: Union[Type[AbstractBackend], AbstractBackend] = AbstractBackend,
    *,
    mangler: Optional[Mangler] = None,
    cachedfactory: Callable[..., Cached] = cachedfactory,
    ttl: int = 3600,
    **backendconf
  ):
    self.ttl = ttl

    mangler = mangler or Mangler()
    assert isinstance(mangler, Mangler)
    self.mangler = mangler

    if isinstance(backend, AbstractBackend):
      if backendconf:
        warnings.warn('Backend options ignored because backend instance is passed')

      self.backend = backend
    elif isinstance(backend, type) and issubclass(backend, AbstractBackend):
      self.backend = backend(self.mangler, **backendconf)
    else:
      raise HermesError('Expected class or instance of AbstractBackend')  # type: ignore

    assert callable(cachedfactory)
    self.cachedfactory = cachedfactory

  def __call__(
    self,
    *args,
    ttl: Optional[int] = None,
    tags: Sequence[str] = (),
    key: Optional[Callable] = None,
  ):
    '''
    Wrap the callable in a cache-point instance.

    Decorator that caches method or function result. The following key
    arguments are optional:

    Bare decorator, ``@cache``, is supported as well as a call with
    keyword arguments ``@cache(ttl = 7200)``.

    :argument ttl:
      Seconds until entry expiration, otherwise instance's default
      is used.
    :argument tags:
      Cache entry tag list.
    :argument key:
      Lambda that provides custom key, otherwise
      :obj:`Mangler.nameEntry` is used.
    '''

    if args:
      # @cache
      return self.cachedfactory(self, args[0])
    else:
      # @cache()
      return functools.partial(self.cachedfactory, self, ttl = ttl, tags = tags, key = key)

  def clean(self, tags: Sequence[str] = ()):
    '''
    Clean all, or tagged with given tags, cache entries.

    :argument tags:
      If this argument is omitted the call flushes all cache entries,
      otherwise only the entries tagged by given tags are flushed.
    '''

    if tags:
      self.backend.remove(map(self.mangler.nameTag, tags))
    else:
      self.backend.clean()


class HermesError(Exception):
  '''Generic Hermes error.'''



import os

from .. import Hermes, test
from ..backend import redis


class TestRedis(test.BaseRemoteBackendTest):

  def setUp(self):
    self.testee = Hermes(
      redis.Backend,
      host = os.getenv('TEST_REDIS_HOST', 'localhost'),
      ttl = 360,
      lockconf = {'timeout': 120},
    )
    super().setUp()

  def getSize(self):
    return self.testee.backend.client.dbsize()

  def getKeyTtl(self, key: str):
    return self.testee.backend.client.ttl(key)

  def testLazyInit(self):
    server = test.FakeBackendServer()
    server.serve()

    Hermes(redis.Backend, port = server.port)

    server.close()
    self.assertEqual([], server.log)


class TestRedisLock(test.BaseRemoteBackendLockTest):

  def setUp(self):
    self.cache = Hermes(redis.Backend, host = os.getenv('TEST_REDIS_HOST', 'localhost'))
    self.testee = redis.Lock('123', self.cache.backend.client)
    super().setUp()

  def getKeyTtl(self, key: str):
    return self.cache.backend.client.ttl(key)



import os
import telnetlib
import time
from typing import Dict

from .. import Hermes, test
from ..backend import memcached


class TestMemcached(test.BaseRemoteBackendTest):

  ttlDelta = 2

  def setUp(self):
    self.testee = Hermes(
      memcached.Backend,
      server = os.getenv('TEST_MEMCACHED_HOST', 'localhost'),
      ttl = 360,
    )
    self.addCleanup(self.testee.backend.client.close)

    super().setUp()

    # force flush memcached
    self.testee.backend.remove(getKeyTtlMapping().keys())

  def getSize(self):
    stats = self.testee.backend.client.stats()
    return int(stats[b'curr_items'])

  def assertEmptyBackend(self):
    # Can not use self.getSize because flushing works even worse in 1.4
    self.assertFalse(self.testee.backend.load(getKeyTtlMapping().keys()))

  def getKeyTtl(self, key: str):
    return max(-1, round(getKeyTtlMapping()[key]))

  def testLazyInit(self):
    server = test.FakeBackendServer()
    server.serve()

    addr = '{}:{}'.format(os.getenv('TEST_MEMCACHED_HOST', 'localhost'), server.port)
    Hermes(memcached.Backend, server = addr)

    server.close()
    self.assertEqual([], server.log)


class TestMemcachedLock(test.BaseRemoteBackendLockTest):

  ttlDelta = 2

  def setUp(self):
    self.cache = Hermes(memcached.Backend, server = os.getenv('TEST_MEMCACHED_HOST', 'localhost'))
    self.addCleanup(self.cache.backend.client.close)

    self.testee = memcached.Lock('123', self.cache.backend.client)

    super().setUp()

    # force flush memcached
    self.cache.backend.remove(getKeyTtlMapping().keys())

  def getKeyTtl(self, key: str):
    return max(-1, round(getKeyTtlMapping()[key]))


def getKeyTtlMapping() -> Dict[str, float]:
  '''
  Get dict that maps key to their TTL.

  What a poorly piece of software! Most of the early issues that happen
  during development are memcached issues. Most notable is that it
  won't flush damn keys. No matter how many times you call it to flush,
  or wait for clean state -- expired by flush call keys are just stale
  with expiry timestamp in past. Current version is 1.4.14.
  '''

  telnet = telnetlib.Telnet(os.getenv('TEST_MEMCACHED_HOST', 'localhost'), 11211)
  now = time.time()
  keys = {}
  try:
    telnet.write(b'stats items\n')
    slablines = telnet.read_until(b'END').split(b'\r\n')
    for slab in {int(s.decode().split(':')[1]) for s in slablines if s.startswith(b'STAT')}:
      telnet.write('stats cachedump {0} 0\n'.format(slab).encode())
      cachelines = telnet.read_until(b'END').split(b'\r\n')
      for line in cachelines:
        parts = line.decode().split(' ', 5)
        if len(parts) < 2:
          continue
        key = parts[1]
        ttl = int(parts[4]) - now
        keys[key] = ttl
  finally:
    telnet.close()

  return keys



import time
from typing import Any, Dict, Iterable, Optional, Union

import pymemcache

from . import AbstractBackend, AbstractLock


__all__ = 'Backend',


class Lock(AbstractLock):
  '''Key-aware distributed lock.'''

  client: pymemcache.PooledClient
  '''Memcached client.'''

  timeout: int
  '''
  Maximum TTL of lock, can be up to 30 days, otherwise memcached will
  treated it as a UNIX timestamp of an exact date.
  '''

  sleep: float
  '''Amount of time to sleep per ``while True`` iteration when waiting.'''

  def __init__(
    self, key: str, client: pymemcache.PooledClient, *, sleep: float = 0.1, timeout: int = 900
  ):
    super().__init__(key)

    self.client  = client
    self.sleep   = sleep
    self.timeout = timeout

  def acquire(self, wait = True):
    while True:
      if self.client.add(self.key, 'locked', self.timeout, noreply = False):
        return True
      elif not wait:
        return False
      else:
        time.sleep(self.sleep)

  def release(self):
    self.client.delete(self.key)


class Backend(AbstractBackend):
  '''Memcached backend implementation.'''

  client: pymemcache.PooledClient
  '''Memcached client.'''

  _lockconf: dict
  '''Lock config.'''

  def __init__(
    self,
    mangler,
    *,
    server: str = 'localhost:11211',
    lockconf: Optional[dict] = None,
    **kwargs,
  ):
    self.mangler = mangler

    # Memcached client creates a pool that connects lazily
    self.client = pymemcache.PooledClient(server, **kwargs)

    self._lockconf = lockconf or {}

  def lock(self, key) -> Lock:
    return Lock(self.mangler.nameLock(key), self.client, **self._lockconf)

  def save(self, mapping: Dict[str, Any], *, ttl: Optional[int] = None):
    mapping = {k: self.mangler.dumps(v) for k, v in mapping.items()}
    self.client.set_multi(mapping, ttl if ttl is not None else 0)

  def load(self, keys: Union[str, Iterable[str]]) -> Optional[Union[Any, Dict[str, Any]]]:
    if isinstance(keys, str):
      value = self.client.get(keys)
      if value is not None:
        value = self.mangler.loads(value)
      return value
    else:
      return {k: self.mangler.loads(v) for k, v in self.client.get_multi(tuple(keys)).items()}

  def remove(self, keys: Union[str, Iterable[str]]):
    if isinstance(keys, str):
      keys = (keys,)

    self.client.delete_multi(keys)

  def clean(self):
    self.client.flush_all()
